"""Foreman provider implementation."""

import inspect
from uuid import uuid4

import click
from dynaconf import Validator
from logzero import logger

from broker.binds import foreman
from broker.helpers import Result
from broker.providers import Provider


class Foreman(Provider):
    """Foreman provider class providing an interface around the Foreman API."""

    _validators = [
        Validator("FOREMAN.organization", must_exist=True),
        Validator("FOREMAN.location", must_exist=True),
        Validator("FOREMAN.foreman_username", must_exist=True),
        Validator("FOREMAN.foreman_password", must_exist=True),
        Validator("FOREMAN.foreman_url", must_exist=True),
        Validator("FOREMAN.hostgroup", must_exist=False),
        Validator("FOREMAN.verify", default=False),
    ]
    _sensitive_attrs = ["foreman_password"]

    _checkout_options = [
        click.option(
            "--hostgroup",
            type=str,
            help="Name of the Foreman hostgroup to deploy the host",
        ),
    ]
    _execute_options = []
    _extend_options = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if kwargs.get("bind") is not None:
            self._runtime_cls = kwargs.pop("bind")
        else:
            self._runtime_cls = foreman.ForemanBind

        self.runtime = self._runtime_cls(
            broker_settings=self._settings,
            foreman_username=kwargs.get(
                "foreman_username", self._settings.foreman.foreman_username
            ),
            foreman_password=kwargs.get(
                "foreman_password", self._settings.foreman.foreman_password
            ),
            url=kwargs.get("url", self._settings.foreman.foreman_url),
            prefix=kwargs.get("prefix", self._settings.foreman.name_prefix),
            verify=kwargs.get("verify", self._settings.foreman.verify),
        )
        self.prefix = self._settings.foreman.name_prefix

        self.organization_id = self.runtime.obtain_id_from_name(
            "organizations", self._settings.foreman.organization
        )
        self.location_id = self.runtime.obtain_id_from_name(
            "locations", self._settings.foreman.location
        )

    def release(self, host):
        """Release a host.

        :param host: Hostname or ID of the host to delete
        """
        data = {
            "organization_id": self.organization_id,
            "location_id": self.location_id,
        }
        self.runtime._delete(
            f"/api/hosts/{host}",
            json=data,
        )

    def extend(self):
        """There is no need to extend a host on Foreman."""
        pass

    def _host_execute(self, command):
        """Execute command on a single host.

        :param command: a command to be executed on the target host

        :return: Result object containing information about the executed job
        """
        hostname = self.last_deployed_host

        return self.execute(
            hostname=hostname,
            job_template="Run Command - Script Default",
            command=command,
        )

    def _parse_job_output(self, job_output):
        """Parse response Foreman gives when querying job output.

        Example data can be found in tests/data/foreman/fake_jobs.json

        Typically the data is a list. In case of failures the last element
        of the output has the element "output_type" "debug" otherwise it is
        "stdout".

        The element before or the last element in case of success contains the
        statusline with the errorcode.
        """
        if job_output[-1]["output_type"] == "debug":
            # command was not successful
            statusline = job_output[-2]["output"]
            status = int(statusline.split(": ")[1])
            stdout = "\n".join([item["output"] for item in job_output[:-2]])
            stderr = ""
        else:
            # command was successful
            status = 0
            stdout = "\n".join([item["output"] for item in job_output[:-1]])
            stderr = ""
        return Result(status=status, stdout=stdout, stderr=stderr)

    def execute(self, hostname, job_template, **kwargs):
        """Execute remote execution on target host.

        :param hostname: hostname to perform remote execution on

        :param job_template: name of the job template

        :param kwargs: input parameters for the job template

        :return: Result object containing all information about executed jobs
        """
        job_template_id = self.runtime.obtain_id_from_name("job_templates", job_template)
        data = {
            "organization_id": self.organization_id,
            "location_id": self.location_id,
            "job_invocation": {
                "job_template_id": job_template_id,
                "targeting_type": "static_query",
                "inputs": kwargs,
                "search_query": f"name={hostname}",
            },
        }
        job_id = self.runtime.create_job_invocation(data)
        self.runtime.wait_for_job_to_finish(job_id)
        job_output = self.runtime.job_output(job_id)

        return self._parse_job_output(job_output)

    def provider_help(
        self,
        hostgroups=False,
        hostgroup=None,
        **kwargs,
    ):
        """Return useful information about Foreman provider."""
        if hostgroups:
            all_hostgroups = self.runtime.hostgroups()
            logger.info(f"On Foreman {self.instance} you have the following hostgroups:")
            for hg in all_hostgroups["results"]:
                logger.info(f"- {hg['title']}")
        elif hostgroup:
            logger.info(
                f"On Foreman {self.instance} the hostgroup {hostgroup} has the following properties:"
            )
            data = self.runtime.hostgroup(name=hostgroup)
            fields_of_interest = {
                "description": "description",
                "operating_system": "operatingsystem_name",
                "domain": "domain_name",
                "subnet": "subnet_name",
                "subnet6": "subnet6_name",
            }
            for name, field in fields_of_interest.items():
                value = data.get(field, False)
                if value:
                    logger.info(f"  {name}: {value}")

    def _compile_host_info(self, host):
        return {
            "name": host["certname"],  # alternatives: name, display_name
            "hostgroup": host["hostgroup_title"],
            "hostname": host["certname"],  # alternatives: name, display_name
            "ip": host["ip"],
            "_broker_provider": "Foreman",
            "_broker_provider_instance": self.instance,
        }

    def get_inventory(self, *args, **kwargs):
        """Synchronize list of hosts on Foreman using set prefix."""
        all_hosts = self.runtime.hosts()
        with click.progressbar(all_hosts, label="Compiling host information") as hosts_bar:
            compiled_host_info = [self._compile_host_info(host) for host in hosts_bar]
        return compiled_host_info

    def _host_release(self):
        """Delete a specific hostDelete a specific host."""
        caller_host = inspect.stack()[1][0].f_locals["host"].hostname
        self.release(caller_host)

    def _set_attributes(self, host_inst, broker_args=None, misc_attrs=None):
        """Extend host object by required parameters and methods."""
        host_inst.__dict__.update(
            {
                "_prov_inst": self,
                "_broker_provider": "Foreman",
                "_broker_args": broker_args,
                "release": self._host_release,
                "execute": self._host_execute,
            }
        )

    def construct_host(self, provider_params, host_classes, **kwargs):
        """Construct a broker host from a Foreman host.

        :param provider_params: a container instance object

        :param host_classes: host object

        :return: broker object of constructed host instance
        """
        logger.debug(f"constructing with {provider_params=}\n{host_classes=}\n{kwargs=}")
        if not provider_params:
            host_inst = host_classes[kwargs.get("type", "host")](**kwargs)
            self._set_attributes(host_inst, broker_args=kwargs)
            return host_inst
        name = provider_params["name"]
        host_inst = host_classes["host"](hostname=name, name=name, broker_settings=self._settings)
        self._set_attributes(host_inst, broker_args=kwargs)
        return host_inst

    def _gen_name(self):
        return f"{self.prefix}-{str(uuid4()).split('-')[0]}"

    @Provider.register_action("hostgroup")
    def create_host(self, hostgroup, **host):
        """Create a new Foreman host.

        :param host: additional parameters for host creation

        :return: Foreman's response from host creation
        """
        if not host.get("name"):
            host["name"] = self._gen_name()

        logger.debug(f"Creating host {host['name']} from hostgroup '{hostgroup}'")

        host["hostgroup_id"] = self.runtime.obtain_id_from_name("hostgroups", hostgroup)
        host["build"] = True
        host["compute_attributes"] = {"start": "1"}
        host["organization_id"] = self.organization_id
        host["location_id"] = self.location_id

        image_name = host.pop("image", False)
        compute_resource_name = host.pop("computeresource", False)
        if image_name and compute_resource_name:
            host["compute_resource_id"] = self.runtime.obtain_id_from_name(
                "compute_resources", compute_resource_name
            )
            host["compute_attributes"]["image_id"] = self.runtime.image_uuid(
                host["compute_resource_id"], image_name
            )
            host["provision_method"] = "image"

            logger.debug(
                "Setting parameters for image based deployment: {\n"
                f"  compute_resource_id: {host['compute_resource_id']},\n"
                f"  image_id: {host['compute_attributes']['image_id']},\n"
                f"  provision_method: {host['provision_method']}\n"
                "}"
            )

        data = {
            "organization_id": self.organization_id,
            "location_id": self.location_id,
            "host": host,
        }
        result = self.runtime.create_host(data)

        self.runtime.wait_for_host_to_install(result["name"])
        self.last_deployed_host = result["name"]

        # set hostname = hostname -f
        self.execute(
            hostname=result["name"],
            job_template="Run Command - Script Default",
            command=f"hostname {result['name']}",
        )
        return result



"""Config migrations for versions older than 0.6.12 to 0.6.12."""

from logzero import logger

TO_VERSION = "0.6.12"


def add_max_resilient_wait_setting(config_dict):
    """Add the max_resilient_wait setting to AnsibleTower provider."""
    logger.debug("Adding max_resilient_wait setting to AnsibleTower provider.")
    ansible_tower = config_dict.get("AnsibleTower", {})
    if "max_resilient_wait" not in ansible_tower:
        ansible_tower["max_resilient_wait"] = 7200  # 2 hours default
    config_dict["AnsibleTower"] = ansible_tower
    return config_dict


def run_migrations(config_dict):
    """Run all migrations."""
    logger.info(f"Running config migrations for {TO_VERSION}.")
    config_dict = add_max_resilient_wait_setting(config_dict)
    config_dict["_version"] = TO_VERSION
    return config_dict



"""Miscellaneous helpers live here."""

import collections
from collections import UserDict, namedtuple
from collections.abc import MutableMapping
from contextlib import contextmanager
from copy import deepcopy
import getpass
import inspect
from io import BytesIO
import json
import os
from pathlib import Path
import sys
import tarfile
import threading
import time
from uuid import uuid4

import click
from logzero import logger
from rich.table import Table
from ruamel.yaml import YAML

from broker import exceptions
from broker.settings import clone_global_settings

FilterTest = namedtuple("FilterTest", "haystack needle test")
INVENTORY_LOCK = threading.Lock()

yaml = YAML()
yaml.default_flow_style = False
yaml.sort_keys = False

SPECIAL_INVENTORY_FIELDS = {}  # use the _special_inventory_field decorator to add new fields


def _special_inventory_field(action_name):
    """Register inventory field actions."""

    def decorator(func):
        SPECIAL_INVENTORY_FIELDS[action_name] = func
        return func

    return decorator


def clean_dict(in_dict):
    """Remove entries from a dict where value is None."""
    return {k: v for k, v in in_dict.items() if v is not None}


def merge_dicts(dict1, dict2):
    """Merge two nested dictionaries together.

    :return: merged dictionary
    """
    if not isinstance(dict1, MutableMapping) or not isinstance(dict2, MutableMapping):
        return dict1
    dict1 = clean_dict(dict1)
    dict2 = clean_dict(dict2)
    merged = {}
    dupe_keys = dict1.keys() & dict2.keys()
    for key in dupe_keys:
        merged[key] = merge_dicts(dict1[key], dict2[key])
    for key in dict1.keys() - dupe_keys:
        merged[key] = deepcopy(dict1[key])
    for key in dict2.keys() - dupe_keys:
        merged[key] = deepcopy(dict2[key])
    return merged


def flatten_dict(nested_dict, parent_key="", separator="_"):
    """Flatten a nested dictionary, keeping nested notation in key.

    {
        'key': 'value1',
        'another': {
            'nested': 'value2',
            'nested2': [1, 2, {'deep': 'value3'}]
        }
    }
    becomes
    {
        "key": "value",
        "another_nested": "value2",
        "another_nested2": [1, 2],
        "another_nested2_deep": "value3"
    }
    note that dictionaries nested in lists will be removed from the list.

    :return: dictionary
    """
    flattened = []
    for key, value in nested_dict.items():
        new_key = f"{parent_key}{separator}{key}" if parent_key else key
        if isinstance(value, dict):
            flattened.extend(flatten_dict(value, new_key, separator).items())
        elif isinstance(value, list):
            to_remove = []
            # avoid mutating nested structures
            value = value.copy()  # noqa: PLW2901
            for index, val in enumerate(value):
                if isinstance(val, dict):
                    flattened.extend(flatten_dict(val, new_key, separator).items())
                    to_remove.append(index)
            for index in to_remove[::-1]:  # remove from back to front
                del value[index]
            flattened.append((new_key, value))
        else:
            flattened.append((new_key, value))
    return dict(flattened)


def dict_from_paths(source_dict, paths, sep="/"):
    """Given a dictionary of desired keys and nested paths, return a new dictionary.

    Example:
        source_dict = {
            "key1": "value1",
            "key2": {
                "nested1": "value2",
                "nested2": {
                    "deep": "value3"
                }
            }
        }
        paths = {
            "key1": "key1",
            "key2": "key2/nested2/deep"
        }
        returns {
            "key1": "value1",
            "key2": "value3"
        }
    """
    result = {}
    for key, path in paths.items():
        if sep not in path:
            result[key] = source_dict.get(path)
        else:
            top, rem = path.split(sep, 1)
            result.update(dict_from_paths(source_dict[top], {key: rem}))
    return result


def eval_filter(filter_list, raw_filter, filter_key="inv"):
    """Run each filter through an eval to get the results."""
    filter_list = [MockStub(item) if isinstance(item, dict) else item for item in filter_list]
    for raw_f in raw_filter.split("|"):
        if f"@{filter_key}[" in raw_f:
            # perform a list filter on the inventory
            filter_list = eval(  # noqa: S307
                raw_f.replace(f"@{filter_key}", filter_key), {filter_key: filter_list}
            )
            filter_list = filter_list if isinstance(filter_list, list) else [filter_list]
        elif f"@{filter_key}" in raw_f:
            # perform an attribute filter on each host
            filter_list = list(
                filter(
                    lambda item: eval(  # noqa: S307
                        raw_f.replace(f"@{filter_key}", filter_key), {filter_key: item}
                    ),
                    filter_list,
                )
            )
    return [dict(item) if isinstance(item, MockStub) else item for item in filter_list]


def resolve_nick(nick, broker_settings=None):
    """Check if the nickname exists. Used to define broker arguments.

    :param nick: String representing the name of a nick
    :param broker_settings: Optional settings object to use instead of global settings

    :return: a dictionary mapping argument names and values
    """
    _settings = broker_settings or clone_global_settings()
    nick_names = _settings.get("NICKS") or {}
    if nick in nick_names:
        return _settings.NICKS[nick].to_dict()
    else:
        raise exceptions.UserError(f"Unknown nick: {nick}")


def load_file(file, warn=True):
    """Verify the existence of and load data from json and yaml files."""
    file = Path(file)
    if not file.exists() or file.suffix not in (".json", ".yaml", ".yml"):
        if warn:
            logger.warning(f"File {file.absolute()} is invalid or does not exist.")
        return []
    if file.suffix == ".json":
        return json.loads(file.read_text())
    elif file.suffix in (".yaml", ".yml"):
        return yaml.load(file)


def resolve_file_args(broker_args):
    """Check for files being passed in as values to arguments then attempt to resolve them.

    If not resolved, keep arg/value pair intact.
    """
    final_args = {}
    # parse the eventual args_file first
    if val := broker_args.pop("args_file", None):
        if isinstance(val, Path) or (isinstance(val, str) and val[-4:] in ("json", "yaml", ".yml")):
            if data := load_file(val):
                if isinstance(data, dict):
                    final_args.update(data)
                elif isinstance(data, list):
                    for d in data:
                        final_args.update(d)
            else:
                raise exceptions.BrokerError(f"No data loaded from {val}")

    for key, val in broker_args.items():
        if isinstance(val, Path) or (isinstance(val, str) and val[-4:] in ("json", "yaml", ".yml")):
            if data := load_file(val):
                final_args.update({key: data})
            else:
                final_args.update({key: val})
        else:
            final_args.update({key: val})
    return final_args


def load_inventory(filter=None):
    """Load all local hosts in inventory.

    :param filter: A filter string to apply to the inventory.

    :return: list of dictionaries
    """
    from broker.settings import inventory_path

    inv_data = load_file(inventory_path, warn=False)
    if inv_data and filter:
        inv_data = eval_filter(inv_data, filter)
    return inv_data or []


def update_inventory(add=None, remove=None):
    """Update list of local hosts in the checkout interface.

    :param add: list of dictionaries representing new hosts
    :param remove: list of strings representing hostnames or names to be removed

    :return: no return value
    """
    from broker.settings import inventory_path

    if add and not isinstance(add, list):
        add = [add]
    elif not add:
        add = []
    if remove and not isinstance(remove, list):
        remove = [remove]
    with INVENTORY_LOCK:
        inv_data = load_inventory()
        if inv_data:
            inventory_path.unlink()

        if remove:
            for host in inv_data[::-1]:
                if host["hostname"] in remove or host.get("name") in remove:
                    # iterate through new hosts and update with old host data if it would nullify
                    for new_host in add:
                        if host["hostname"] == new_host["hostname"] or host.get(
                            "name"
                        ) == new_host.get("name"):
                            # update missing data in the new_host with the old_host data
                            new_host.update(merge_dicts(new_host, host))
                    inv_data.remove(host)
        if add:
            inv_data.extend(add)

        inventory_path.touch()
        yaml.dump(inv_data, inventory_path)


def yaml_format(in_struct, force_yaml_dict=False):
    """Convert a yaml-compatible structure to a yaml dumped string.

    :param in_struct: yaml-compatible structure or string containing structure
    :param force_yaml_dict: force the in_struct to be converted to a dictionary before dumping

    :return: yaml-formatted string
    """
    if isinstance(in_struct, str):
        # first try to load is as json
        try:
            in_struct = json.loads(in_struct)
        except json.JSONDecodeError:
            # then try yaml
            in_struct = yaml.load(in_struct)
            if force_yaml_dict:
                in_struct = dict(in_struct)
    output = BytesIO()  # ruamel doesn't natively allow for string output
    yaml.dump(in_struct, output)
    return output.getvalue().decode("utf-8")


def flip_provider_actions(provider_actions):
    """Flip the mapping of actions->provider to provider->actions."""
    flipped = {}
    for action, (provider, _) in provider_actions.items():
        provider_name = provider.__name__
        if provider_name not in flipped:
            flipped[provider_name] = []
        flipped[provider_name].append(action)
    return flipped


def inventory_fields_to_dict(inventory_fields, host_dict, **extras):
    """Convert a dicionary-like representation of inventory fields to a resolved dictionary.

    inventory fields, as set in the config look like this, in yaml:
    inventory_fields:
        Host: hostname | name
        Provider: _broker_provider
        Action: $action
        OS: os_distribution os_distribution_version

    We then process that into a dictionary with inventory values like this:
    {
        "Host": "some.test.host",
        "Provider": "AnsibleTower",
        "Action": "deploy-rhel",
        "OS": "RHEL 8.4"
    }

    Notes: The special syntax use in Host and Action fields <$action> is a special keyword that
    represents a more complex field resolved by Broker.
    Also, the Host field represents a priority  order of single values,
    so if hostname is not present, name will be used.
    Finally, spaces between values are preserved. This lets us combine multiple values in a single field.
    """
    return {
        name: _resolve_inv_field(field, host_dict, **extras)
        for name, field in inventory_fields.items()
    }


def _resolve_inv_field(field, host_dict, **extras):
    """Real functionality for inventory_fields_to_dict, allows recursive evaluation."""
    # Users can specify multiple values to try in order of priority, so evaluate each
    if "|" in field:
        resolved = [_resolve_inv_field(f.strip(), host_dict, **extras) for f in field.split("|")]
        for val in resolved:
            if val and val != "Unknown":
                return val
        return "Unknown"
    # Users can combine multiple values in a single field, so evaluate each
    if " " in field:
        return " ".join(_resolve_inv_field(f, host_dict, **extras) for f in field.split())
    # Some field values require special handling beyond what the existing syntax allows
    if special_field_func := SPECIAL_INVENTORY_FIELDS.get(field):
        return special_field_func(host_dict, **extras)
    # Otherwise, try to get the value from the host dictionary
    return dict_from_paths(host_dict, {"_": field}, sep=".")["_"] or "Unknown"


@_special_inventory_field("$action")
def get_host_action(host_dict, provider_actions=None, **_):
    """Get a more focused set of fields from the host inventory."""
    if not provider_actions:
        return "$actionError"
    # Flip the mapping of actions->provider to provider->actions
    flipped_actions = {}
    for action, (provider, _) in provider_actions.items():
        provider_name = provider.__name__
        if provider_name not in flipped_actions:
            flipped_actions[provider_name] = []
        flipped_actions[provider_name].append(action)
    # Get the host's action, based on its provider
    provider = host_dict["_broker_provider"]
    for opt in flipped_actions[provider]:
        if action := host_dict["_broker_args"].get(opt):
            return action
    return "Unknown"


def kwargs_from_click_ctx(ctx):
    """Convert a Click context object to a dictionary of keyword arguments."""
    # if users use `=` to note arg=value assignment, then we need to split it
    _args = []
    for arg in ctx.args:
        if "=" in arg:
            _args.extend(arg.split("="))
        else:
            _args.append(arg)
    ctx.args = _args
    # if additional arguments were passed, include them in the broker args
    # strip leading -- characters
    return {
        (key[2:] if key.startswith("--") else key): val
        for key, val in zip(ctx.args[::2], ctx.args[1::2])
    }


class Emitter:
    """Class that provides a simple interface to emit messages to a json-formatted file.

    This module also has an instance of this class called "emit" that should be used
    instead of this class directly.

    Usage examples:
        helpers.emit(key=value, another=5)
        helpers.emit({"key": "value", "another": 5})
    """

    EMIT_LOCK = threading.Lock()

    def __init__(self, emit_file=None):
        """Can empty init and set the file later."""
        self.file = None
        if emit_file:
            self.file = self.set_file(emit_file)

    def set_file(self, file_path):
        """Set the file to emit to."""
        if file_path:
            self.file = Path(file_path)
            self.file.parent.mkdir(exist_ok=True, parents=True)
            if self.file.exists():
                self.file.unlink()
            self.file.touch()

    def emit_to_file(self, *args, **kwargs):
        """Emit data to the file, keeping existing data in-place."""
        if not self.file:
            return
        for arg in args:
            if not isinstance(arg, dict):
                raise exceptions.BrokerError(f"Received an invalid data emission {arg}")
            kwargs.update(arg)
        for key, val in kwargs.items():
            if getattr(val, "json", None):
                kwargs[key] = val.json
        with self.EMIT_LOCK:
            curr_data = json.loads(self.file.read_text() or "{}")
            curr_data.update(kwargs)
            self.file.write_text(json.dumps(curr_data, indent=4, sort_keys=True))

    def __call__(self, *args, **kwargs):
        """Allow emit to be used like a function."""
        return self.emit_to_file(*args, **kwargs)


emit = Emitter()


class MockStub(UserDict):
    """Test helper class. Allows for both arbitrary mocking and stubbing."""

    def __init__(self, in_dict=None):
        """Initialize the class and all nested dictionaries."""
        if in_dict is None:
            in_dict = {}
        for key, value in in_dict.items():
            if isinstance(value, dict):
                setattr(self, key, MockStub(value))
            elif type(value) in (list, tuple):
                setattr(
                    self,
                    key,
                    [MockStub(x) if isinstance(x, dict) else x for x in value],
                )
            else:
                setattr(self, key, value)
        super().__init__(in_dict)

    def __getattr__(self, name):
        """Fallback to returning self if attribute doesn't exist."""
        return self

    def __getitem__(self, key):
        """Get an item from the dictionary-like object.

        If the key is a string, this method will attempt to get an attribute with that name.
        If the key is not found, this method will return the object itself.
        """
        if isinstance(key, str):
            item = getattr(self, key, self)
        try:
            item = super().__getitem__(key)
        except KeyError:
            item = self
        return item

    def __call__(self, *args, **kwargs):
        """Allow MockStub to be used like a function."""
        return self

    def __hash__(self):
        """Return a hash value for the object.

        The hash value is computed using the hash value of all hashable attributes of the object.
        """
        return hash(
            tuple(kp for kp in self.__dict__.items() if isinstance(kp[1], collections.abc.Hashable))
        )


def update_log_level(ctx, param, value):
    """Update the log level and file logging settings for the Broker.

    Args:
        ctx: The Click context object.
        param: The Click parameter object.
        value: The new log level value.
    """
    from broker import logger as b_log

    b_log.set_log_level(value)
    b_log.set_file_logging(value)


def set_emit_file(ctx, param, value):
    """Update the file that the Broker emits data to."""
    emit.set_file(value)


def fork_broker():
    """Fork the Broker process to run in the background."""
    pid = os.fork()
    if pid:
        logger.info(f"Running broker in the background with pid: {pid}")
        sys.exit(0)
    from broker import logger as b_log

    b_log.set_log_level("silent")
    b_log.set_file_logging("silent")


def handle_keyboardinterrupt(*args):
    """Handle keyboard interrupts gracefully.

    Offer the user a choice between keeping Broker alive in the background, killing it, or resuming execution.
    """
    choice = click.prompt(
        "\nEnding Broker while running may not end processes being monitored.\n"
        "Would you like to switch Broker to run in the Background, Kill it, or Resume execution?\n",
        type=click.Choice(["b", "k", "r"]),
        default="r",
    ).lower()
    if choice == "b":
        fork_broker()
    elif choice == "k":
        raise exceptions.BrokerError("Broker killed by user.")
    elif choice == "r":
        click.echo("Resuming execution...")


def translate_timeout(timeout):
    """Allow for flexible timeout definitions, converts other units to ms.

    acceptable units are (s)econds, (m)inutes, (h)ours, (d)ays
    """
    if isinstance(timeout, str):
        timeout, unit = int(timeout[:-1]), timeout[-1]
        if unit == "d":
            timeout *= 24
            unit = "h"
        if unit == "h":
            timeout *= 60
            unit = "m"
        if unit == "m":
            timeout *= 60
            unit = "s"
        if unit == "s":
            timeout *= 1000
    return timeout if isinstance(timeout, int) else 0


def simple_retry(
    cmd, cmd_args=None, cmd_kwargs=None, max_timeout=60, _cur_timeout=1, terminal_exceptions=None
):
    """Re(Try) a function given its args and kwargs up until a max timeout."""
    cmd_args = cmd_args if cmd_args else []
    cmd_kwargs = cmd_kwargs if cmd_kwargs else {}
    terminal_exceptions = terminal_exceptions or ()

    try:
        return cmd(*cmd_args, **cmd_kwargs)
    except terminal_exceptions:
        raise
    except Exception as err:
        new_wait = _cur_timeout * 2
        if new_wait > max_timeout:
            raise err
        logger.warning(
            f"Tried {cmd=} with {cmd_args=}, {cmd_kwargs=} but received {err=}"
            f"\nTrying again in {_cur_timeout} seconds."
        )
        time.sleep(_cur_timeout)
        simple_retry(cmd, cmd_args, cmd_kwargs, max_timeout, new_wait, terminal_exceptions)


class FileLock:
    """Basic file locking class that acquires and releases locks.

    Recommended usage is the context manager which will handle everything for you

    with FileLock("basic_file.txt"):
        Path("basic_file.txt").write_text("some text")

    If a lock is already in place, FileLock will wait up to <timeout> seconds
    """

    def __init__(self, file_name, timeout=10):
        self.lock = Path(f"{file_name}.lock")
        self.timeout = timeout

    def wait_file(self):
        """Wait for the lock file to be released, then acquire it."""
        timeout_after = time.time() + self.timeout
        while self.lock.exists():
            if time.time() <= timeout_after:
                time.sleep(1)
            else:
                raise exceptions.BrokerError(
                    f"Timeout while waiting for lock release: {self.lock.absolute()}"
                )
        self.lock.touch()

    def return_file(self):
        """Release the lock file."""
        self.lock.unlink()

    def __enter__(self):  # noqa: D105
        self.wait_file()

    def __exit__(self, *tb_info):  # noqa: D105
        self.return_file()


class Result:
    """Dummy result class for presenting results in dot access."""

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        """Return a string representation of the object."""
        return f"stdout:\n{self.stdout}\nstderr:\n{self.stderr}\nstatus: {self.status}"

    @classmethod
    def from_ssh(cls, stdout, channel):
        """Create a Result object from an SSH channel."""
        return cls(
            stdout=stdout,
            status=channel.get_exit_status(),
            stderr=channel.read_stderr()[1].decode("utf-8"),
        )

    @classmethod
    def from_duplexed_exec(cls, duplex_exec, runtime=None):
        """Create a Result object from a duplexed exec object from podman or docker."""
        if runtime == "podman":
            status, (stdout, stderr) = duplex_exec
            return cls(
                status=status,
                stdout=stdout.decode("utf-8") if stdout else "",
                stderr=stderr.decode("utf-8") if stderr else "",
            )

        if duplex_exec.output[0]:
            stdout = duplex_exec.output[0].decode("utf-8")
        else:
            stdout = ""
        if duplex_exec.output[1]:
            stderr = duplex_exec.output[1].decode("utf-8")
        else:
            stderr = ""
        return cls(
            status=duplex_exec.exit_code,
            stdout=stdout,
            stderr=stderr,
        )

    @classmethod
    def from_nonduplexed_exec(cls, nonduplex_exec):
        """Create a Result object from a nonduplexed exec object from the docker library."""
        return cls(
            status=nonduplex_exec.exit_code,
            stdout=nonduplex_exec.output.decode("utf-8"),
            stderr="",
        )


def find_origin():
    """Move up the call stack to find tests, fixtures, or cli invocations.

    Additionally, return the jenkins url, if it exists.
    """
    prev, jenkins_url = None, os.environ.get("BUILD_URL")
    for frame in inspect.stack():
        if frame.function == "checkout" and frame.filename.endswith("broker/commands.py"):
            return f"broker_cli:{getpass.getuser()}", jenkins_url
        if frame.function.startswith("test_"):
            return f"{frame.function}:{frame.filename}", jenkins_url
        if frame.function == "call_fixture_func":
            # attempt to find the test name from the fixture's request object
            if request := _frame.frame.f_locals.get("request"):  # noqa: F821
                return f"{prev} for {request.node._nodeid}", jenkins_url
            # otherwise, return the fixture name and filename
            return prev or "Uknown fixture", jenkins_url
        prev, _frame = f"{frame.function}:{frame.filename}", frame
    return f"Unknown origin by {getpass.getuser()}", jenkins_url


@contextmanager
def data_to_tempfile(data, path=None, as_tar=False):
    """Write data to a temporary file and return the path."""
    path = Path(path or uuid4().hex[-10])
    logger.debug(f"Creating temporary file {path.absolute()}")
    if isinstance(data, bytes):
        path.write_bytes(data)
    elif isinstance(data, str):
        path.write_text(data)
    else:
        raise TypeError(f"data must be bytes or str, not {type(data)}")
    if as_tar:
        tar = tarfile.open(path)
        yield tarfile.open(path)
        tar.close()
    else:
        yield path
    path.unlink()


@contextmanager
def temporary_tar(paths):
    """Create a temporary tar file and return the path."""
    temp_tar = Path(f"{uuid4().hex[-10]}.tar")
    with tarfile.open(temp_tar, mode="w") as tar:
        for path in paths:
            logger.debug(f"Adding {path.absolute()} to {temp_tar.absolute()}")
            tar.add(path, arcname=path.name)
    yield temp_tar.absolute()
    temp_tar.unlink()


def dictlist_to_table(dict_list, title=None, _id=False, headers=True):
    """Convert a list of dictionaries to a rich table."""
    # I like pretty colors, so let's cycle through them
    column_colors = ["cyan", "magenta", "green", "yellow", "blue", "red"]
    curr_color = 0
    table = Table(title=title)
    # construct the columns
    if _id:  # likely just for inventory tables
        table.add_column("Id", justify="left", style=column_colors[curr_color], no_wrap=True)
        curr_color += 1
    for key in dict_list[0]:  # assume all dicts have the same keys
        table.add_column(key, justify="left", style=column_colors[curr_color])
        curr_color += 1
        if curr_color >= len(column_colors):
            curr_color = 0
    # add the rows
    for id_num, data_dict in enumerate(dict_list):
        row = [str(id_num)] if _id else []
        row.extend([str(value) for value in data_dict.values()])
        table.add_row(*row)
    if not headers:
        table.show_header = False
    return table


def dict_to_table(in_dict, title=None, headers=None):
    """Convert a dictionary into a rich table."""
    # need to normalize the values first
    in_dict = {k: str(v) for k, v in in_dict.items()}
    table = Table(title=title)
    if isinstance(headers, tuple | list):
        table.add_column(headers[0], style="cyan")
        table.add_column(headers[1], style="magenta")
    else:
        table.add_column("key", style="cyan")
        table.add_column("value", style="magenta")
        table.show_header = False
    for key, val in in_dict.items():
        table.add_row(key, val)
    return table



"""Container provider implementation."""

from functools import cache
import getpass
import inspect
from uuid import uuid4

import click
from dynaconf import Validator
from logzero import logger

from broker import exceptions, helpers
from broker.binds import containers
from broker.providers import Provider


def container_info(container_inst):
    """Return a dict of container information."""
    attr_dict = {"container_host": "Image", "_broker_origin": "Labels/broker.origin"}
    info = {
        "_broker_provider": "Container",
        "_broker_args": helpers.dict_from_paths(container_inst.attrs, attr_dict),
        "name": container_inst.name,
        "hostname": container_inst.attrs["Config"].get("Hostname"),
        "image": container_inst.image.tags,
        "ports": container_inst.ports,
    }
    try:
        info["status"] = container_inst.status
    except TypeError:
        info["status"] = container_inst.attrs["State"]
    return info


def _host_release():
    caller_host = inspect.stack()[1][0].f_locals["host"]
    if not caller_host._cont_inst:
        caller_host._cont_inst = caller_host._prov_inst._cont_inst_by_name(caller_host.name)
    caller_host._cont_inst.remove(v=True, force=True)
    caller_host._checked_in = True


@cache
def get_runtime(
    runtime_cls=None,
    host=None,
    username=None,
    password=None,
    port=None,
    timeout=None,
    broker_settings=None,
):
    """Return a runtime instance."""
    return runtime_cls(host=host, username=username, password=password, port=port, timeout=timeout)


class Container(Provider):
    """Container provider class providing a Broker interface around the container binds."""

    _validators = [
        Validator("CONTAINER.runtime", default="podman"),
        Validator("CONTAINER.host", default="localhost"),
        Validator("CONTAINER.host_username", default="root"),
        Validator(
            "CONTAINER.host_password",
        ),
        Validator("CONTAINER.host_port", default=2375),
        Validator("CONTAINER.network", default=None),
        Validator("CONTAINER.timeout", default=360),
        Validator("CONTAINER.auto_map_ports", is_type_of=bool, default=True),
    ]
    _checkout_options = [
        click.option(
            "--container-host",
            type=str,
            help="Name of a broker-compatible container host image",
        ),
    ]
    _execute_options = [
        click.option(
            "--container-app",
            type=str,
            help="Name of a container application image",
        ),
    ]
    _extend_options = []

    _sensitive_attrs = ["password", "host_password"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if kwargs.get("bind") is not None:
            self._runtime_cls = kwargs.pop("bind")
        elif self._settings.Container.runtime.lower() == "podman":
            self._runtime_cls = containers.PodmanBind
        elif self._settings.Container.runtime.lower() == "docker":
            self._runtime_cls = containers.DockerBind
        else:
            raise exceptions.ProviderError(
                "Container",
                f"Broker has no bind for {self._settings.Container.runtime} containers",
            )
        self.runtime = get_runtime(
            runtime_cls=self._runtime_cls,
            host=self._settings.Container.host,
            username=self._settings.Container.host_username,
            password=self._settings.Container.host_password,
            port=self._settings.Container.host_port,
            timeout=self._settings.Container.timeout,
            broker_settings=self._settings,
        )
        self._name_prefix = self._settings.Container.get("name_prefix", getpass.getuser())

    def _ensure_image(self, name):
        """Check if an image exists on the provider, attempt a pull if not."""
        for image in self.runtime.images:
            if name in image.tags:
                return
            elif ("localhost/" in name) and (name[10:] in image.tags):
                return
        try:
            self.runtime.pull_image(name)
        except Exception as err:
            raise exceptions.ProviderError(
                "Container", f"Unable to find image: {name}\n{err}"
            ) from err

    @staticmethod
    def _find_ssh_port(port_map, ssh_port=22):
        """Go through container port map and find the mapping that corresponds to port 22."""
        if isinstance(port_map, list):
            # [{'hostPort': 1337, 'containerPort': 22, 'protocol': 'tcp', 'hostIP': ''},
            for pm in port_map:
                if pm["containerPort"] == ssh_port:
                    return pm["hostPort"]
        elif isinstance(port_map, dict):
            # {'22/tcp': [{'HostIp': '', 'HostPort': '1337'}],
            for key, val in port_map.items():
                if key.startswith("22") and isinstance(val, list):
                    return val[0]["HostPort"]

    def _set_attributes(self, host_inst, broker_args=None, cont_inst=None):
        host_inst.__dict__.update(
            {
                "_prov_inst": self,
                "_cont_inst": cont_inst,
                "_broker_provider": "Container",
                "_broker_provider_instance": self.instance,
                "_broker_args": broker_args,
                "release": _host_release,
            }
        )

    def _port_mapping(self, image, **kwargs):
        """Create a mapping of ports to expose on the container.

        Accepted `ports` formats:
            22
            22:1337
            22/tcp
            22/tcp:1337
            22,23
            22:1337 23:1335.
        """
        mapping = {}
        # create mapping for all exposed ports in the image
        if self._settings.Container.auto_map_ports or kwargs.get("auto_map_ports"):
            mapping = {
                k: v or None
                for k, v in self.runtime.image_info(image)["config"]["ExposedPorts"].items()
            }
        # add any ports that were passed as arguments
        if ports := kwargs.pop("ports", None):
            if isinstance(ports, str):
                for _map in ports.split():
                    if ":" in _map:
                        p, h = _map.split(":")
                    else:
                        p, h = _map, None
                    if "/" in p:
                        p, s = p.split("/")
                    else:
                        s = "tcp"
                    mapping[f"{p}/{s}"] = int(h) if h else None
        return mapping

    def _cont_inst_by_name(self, cont_name):
        """Attempt to find and return a container by its name."""
        for cont_inst in self.runtime.containers:
            if cont_inst.name == cont_name:
                return cont_inst
        logger.error(f"Unable to find container by name {cont_name}")

    def construct_host(self, provider_params, host_classes, **kwargs):
        """Construct a broker host from a container instance.

        :param provider_params: a container instance object

        :param host_classes: host object

        :return: broker object of constructed host instance
        """
        logger.debug(f"constructing with {provider_params=}\n{host_classes=}\n{kwargs=}")
        if not provider_params:
            host_inst = host_classes[kwargs.get("type", "host")](**kwargs)
            cont_inst = self._cont_inst_by_name(host_inst.name)
            self._set_attributes(host_inst, broker_args=kwargs, cont_inst=cont_inst)
            return host_inst
        cont_inst = provider_params
        cont_attrs = self.runtime.get_attrs(cont_inst)
        logger.debug(cont_attrs)
        hostname = cont_inst.attrs["Config"].get("Hostname")
        if port := self._find_ssh_port(cont_attrs["ports"]):
            hostname = f"{self.runtime.host}:{port}"
        if not hostname:
            raise Exception(f"Could not determine container hostname:\n{cont_attrs}")
        name = cont_attrs["name"]
        logger.debug(f"hostname: {hostname}, name: {name}, host type: host")
        host_inst = host_classes["host"](
            hostname=hostname, name=name, broker_settings=self._settings
        )
        self._set_attributes(host_inst, broker_args=kwargs, cont_inst=cont_inst)
        # add the container's port mapping to the host instance only if there are any ports open
        if cont_attrs.get("ports"):
            host_inst.exposed_ports = {
                f"{k.split('/')[0]}": v[0]["HostPort"] for k, v in cont_attrs["ports"].items() if v
            }
        return host_inst

    def provider_help(
        self, container_hosts=False, container_host=None, container_apps=False, **kwargs
    ):
        """Return useful information about container images."""
        results_limit = kwargs.get("results_limit", self._settings.Container.results_limit)
        if container_host:
            logger.info(
                f"Information for {container_host} container-host:\n"
                f"{helpers.yaml_format(self.runtime.image_info(container_host))}"
            )
        elif container_hosts:
            images = [
                img.tags[0]
                for img in self.runtime.images
                if img.labels.get("broker_compatible") and img.tags
            ]
            if res_filter := kwargs.get("results_filter"):
                images = helpers.eval_filter(images, res_filter, "res")
                images = images if isinstance(images, list) else [images]
            images = "\n".join(images[:results_limit])
            logger.info(f"Available host images:\n{images}")
        elif container_apps:
            images = [img.tags[0] for img in self.runtime.images if img.tags]
            if res_filter := kwargs.get("results_filter"):
                images = helpers.eval_filter(images, res_filter, "res")
                images = images if isinstance(images, list) else [images]
            images = "\n".join(images[:results_limit])
            logger.info(f"Available app images:\n{images}")

    def get_inventory(self, name_prefix):
        """Get all containers that have a matching name prefix."""
        name_prefix = name_prefix or self._name_prefix
        return [
            container_info(cont)
            for cont in self.runtime.containers
            if cont.name.startswith(name_prefix)
        ]

    def extend(self):
        """There is no need to extend a continer-ased host."""

    def release(self, host_obj):
        """Remove a container-based host from the container host."""
        host_obj._cont_inst.remove(force=True)

    @Provider.register_action("container_host")
    def run_container(self, container_host, **kwargs):
        """Start a container based on an image name (container_host)."""
        self._ensure_image(container_host)
        if not kwargs.get("name"):
            kwargs["name"] = self._gen_name()
        kwargs["ports"] = self._port_mapping(container_host, **kwargs)

        # Handle network configuration
        if net_name := getattr(self._settings.Container, "network", None):
            net_dict = {}
            for name in net_name.split(","):
                # Note: In real implementation, you might want to check if network exists
                net_dict[name] = {"NetworkId": name}
            kwargs["networks"] = net_dict

        envars = kwargs.get("environment", {})
        if isinstance(envars, str):
            envars = {var.split("=")[0]: var.split("=")[1] for var in envars.split(",")}
        # add some context information about the container's requester
        origin = helpers.find_origin()

        if "for" in origin:
            origin = origin.split()[-1]
        envars["BROKER_ORIGIN"] = origin[0]
        if origin[1]:
            envars["JENKINS_URL"] = origin[1]
        kwargs["environment"] = envars

        # process eventual provider labels for each setting level
        kwargs["provider_labels"] = kwargs.get("provider_labels", {})
        kwargs["provider_labels"].update(self._settings.get("provider_labels", {}))
        kwargs["provider_labels"].update(self._settings.Container.get("provider_labels", {}))
        # prefix eventual label keys with 'broker.' to conform to the docker guidelines
        # https://docs.docker.com/config/labels-custom-metadata/#key-format-recommendations
        kwargs["provider_labels"] = {
            f"broker.{label[0]}": label[1] for label in kwargs.get("provider_labels", {}).items()
        }
        # process eventual labels that were passed externally, split by "="
        kwargs["provider_labels"].update(
            {"broker.origin": origin[0], "broker.jenkins.url": origin[1]}
        )
        # rename the dict key to the name of the arg recognized by provider
        kwargs["labels"] = kwargs.pop("provider_labels")
        container_inst = self.runtime.create_container(
            container_host, net_name=self._settings.container.network, **kwargs
        )
        container_inst.start()
        return container_inst

    @Provider.register_action("container_app")
    def execute(self, container_app, **kwargs):
        """Run a container and return the raw results."""
        if not kwargs.get("name"):
            kwargs["name"] = self._gen_name()
        return self.runtime.execute(container_app, **kwargs)

    def run_wait_container(self, image_name, **kwargs):
        """Run a container and wait for it to exit."""
        cont_inst = self.run_container(image_name, **kwargs)
        cont_inst.wait(condition="exited")
        return self.runtime.get_logs(cont_inst)

    def _gen_name(self):
        return f"{self._name_prefix}_{str(uuid4()).split('-')[0]}"



"""Config migrations for versions older than 0.6.1 to 0.6.1.

Copy this file to a new file in the same directory and modify it to create a new migration.
The new file must be named `vX_Y_Z.py` where X_Y_Z is the version you are migrating to.

e.g. cp example_migration.py v0_6_1.py
"""

from logzero import logger

TO_VERSION = "0.6.1"


def example_migration(config_dict):
    """Migrations should modify the config_dict in place and return it."""
    config_dict["example_key"] = "example_value"
    return config_dict


def run_migrations(config_dict):
    """Run all migrations."""
    logger.info(f"Running config migrations for {TO_VERSION}.")
    config_dict = example_migration(config_dict)
    config_dict["_version"] = TO_VERSION
    return config_dict



"""Ansible Tower provider implementation."""

from functools import cache, cached_property
import inspect
import json
import os
import random
import string
import sys
import time
from urllib import parse as url_parser

import click
from dynaconf import Validator
from logzero import logger
from packaging.version import InvalidVersion, Version
from requests.exceptions import ConnectionError
from rich.console import Console
from rich.prompt import Prompt

from broker import exceptions, helpers
from broker.helpers import MockStub, eval_filter, find_origin, update_inventory, yaml
from broker.providers import Provider
from broker.settings import clone_global_settings

# Module-level variables for deferred imports
awxkit = None
AAP_URL_PREFIX = "#"  # Used to construct the UI URL


def detect_and_reconfigure_for_aap25(base_url, broker_settings=None):
    """Detect AAP 2.5+ and reconfigure awxkit if needed.

    This function can be called multiple times safely.
    Returns True if reconfiguration was done, False otherwise.
    """
    global awxkit, AAP_URL_PREFIX  # noqa: PLW0603

    _settings = broker_settings or clone_global_settings()
    # Exit immediately if we've already configured
    if getattr(detect_and_reconfigure_for_aap25, "_configured", False):
        return False

    # Check to see if the user wants us to operate against AAP 2.4-
    need_reconfigure = False
    try:
        if Version(_settings.ANSIBLETOWER.get("AAP_VERSION")) < Version("2.5"):
            return False
        need_reconfigure = True
    except (InvalidVersion, TypeError):
        pass  # aap_version isn't set or is invalid, need to manually check

    # Only perform API check if we don't already know we need to reconfigure
    if not need_reconfigure:
        config = awxkit.config
        config.base_url = base_url
        try:
            # Create temporary API instance to test version
            api = awxkit.api.Api()
            api.get().available_versions
        except AttributeError:
            # AAP 2.5+ hits this error, so we need to reconfigure
            logger.debug("AAP 2.5+ detected, need to reconfigure awxkit")
            need_reconfigure = True

    # Perform reconfiguration only if needed
    if need_reconfigure:
        logger.debug("Reconfiguring awxkit for AAP 2.5+")

        # Remove all awxkit modules from sys.modules
        for module_name in list(sys.modules.keys()):
            if "awx" in module_name:
                del sys.modules[module_name]

        # Set environment variable for API path
        os.environ["AWXKIT_API_BASE_PATH"] = "/api/controller/"

        # Re-import awxkit
        import awxkit

        AAP_URL_PREFIX = "execution"

        # Configure ruamel yaml representer for PseudoNamespace
        yaml.representer.add_representer(
            awxkit.utils.PseudoNamespace,
            lambda dumper, data: dumper.represent_dict(dict(data)),
        )

        # Set flag to avoid redundant reconfiguration
        detect_and_reconfigure_for_aap25._configured = True
        return True

    return False


def convert_pseudonamespaces(attr_dict):
    """Recursively convert PseudoNamespace objects into dictionaries."""
    out_dict = {}
    for key, value in attr_dict.items():
        if isinstance(value, awxkit.utils.PseudoNamespace):
            out_dict[key] = dict(value)
        elif isinstance(value, dict):
            out_dict[key] = convert_pseudonamespaces(value)
        else:
            out_dict[key] = value
    return out_dict


def resilient_job_wait(job, job_timeout=None, max_wait=None, broker_settings=None):
    """Wait for a job to complete. Retry on errors.

    Args:
        job: The job object to wait for
        job_timeout: Timeout for individual job wait attempts
        max_wait: Maximum time to continue retrying before giving up
        broker_settings: Optional settings object to use instead of the global one
    """
    _settings = broker_settings or clone_global_settings()
    job_timeout = job_timeout or _settings.ANSIBLETOWER.workflow_timeout
    max_wait = max_wait or _settings.ANSIBLETOWER.max_resilient_wait
    completed = False
    start_time = time.time()

    while not completed:
        # Check if we've exceeded max wait time
        if time.time() - start_time > max_wait:
            raise JobExecutionError(
                message_data={
                    "error": "Maximum resilient wait time exceeded",
                    "max_wait_seconds": max_wait,
                    "elapsed_seconds": int(time.time() - start_time),
                }
            )

        try:
            job.wait_until_completed(timeout=job_timeout)
            completed = True
        except (ConnectionError, awxkit.exceptions.Unknown, awxkit.exceptions.Forbidden) as err:
            logger.error(f"Error occurred while waiting for job: {err}")
            logger.info("Retrying job wait...")
            # Add a small delay before retrying to avoid hammering the API
            time.sleep(5)


class JobExecutionError(exceptions.ProviderError):
    """Raised when a job execution fails."""

    def __init__(self, message_data=None):
        super().__init__(
            provider="AnsibleTower",
            message=json.dumps(message_data, indent=2),
        )


class ATInventoryError(exceptions.ProviderError):
    """Raised when we can't find the right inventory."""

    def __init__(self, message=None):
        super().__init__(
            provider="AnsibleTower",
            message=message,
        )


@cache
def get_awxkit_and_uname(
    awxkit_config=None,
    root=None,
    url=None,
    token=None,
    uname=None,
    pword=None,
    broker_settings=None,
):
    """Return an awxkit api object and resolved username."""
    _settings = broker_settings or clone_global_settings()
    if not isinstance(awxkit_config, MockStub):  # skip if we're in a unit test
        # Configure for AAP 2.5+ if needed, with the URL we have
        detect_and_reconfigure_for_aap25(base_url=url, broker_settings=_settings)

    # Now proceed with the original function logic
    awxkit_config = awxkit_config or awxkit.config
    awxkit_config.base_url = url
    if root is None:  # support mock stub for unit tests
        root = awxkit.api.Api()

    temp_token_desc = None

    # If no token was provided, try to create a temporary token
    if not token and uname and pword:
        helpers.emit(auth_type="password")
        logger.warning(
            "You should be using token-based authentication.\n"
            "I will attempt to create and use a temporary token."
        )

        # Set up password-based auth temporarily to create a token
        awxkit_config.credentials = {"default": {"username": uname, "password": pword}}
        awxkit_config.use_sessions = True
        root.load_session().get()
        versions = root.available_versions

        # Create temporary token with random description
        try:
            temp_token_desc = "Broker temp " + "".join(random.choices(string.ascii_letters, k=10))
            if token := root.get_oauth2_token(description=temp_token_desc):
                logger.info(
                    f"Successfully created a temporary token with description: {temp_token_desc}."
                )
        except Exception as err:  # noqa: BLE001 - Don't currently know the specific exception
            logger.debug(f"Error creating temporary token: {err}")
            logger.warning(
                "Failed to create temporary token. Continuing with password authentication"
            )

    # Now proceed with token authentication (either provided or newly created)
    if token:
        helpers.emit(auth_type="token")
        logger.debug("Using token authentication")
        awxkit_config.token = token
        try:
            root.connection.login(username=None, password=None, token=token, auth_type="Bearer")
        except awxkit.exceptions.Unauthorized as err:
            raise exceptions.AuthenticationError(err.args[0]) from err
        versions = root.get().available_versions

    return versions.v2.get(), temp_token_desc


class AnsibleTower(Provider):
    """Ansible Tower provider provides a Broker-specific wrapper around awxkit."""

    _validators = [
        Validator("ANSIBLETOWER.release_workflow", default="remove-vm"),
        Validator("ANSIBLETOWER.extend_workflow", default="extend-vm"),
        Validator("ANSIBLETOWER.new_expire_time", default="+172800"),
        Validator("ANSIBLETOWER.workflow_timeout", is_type_of=int, default=3600),
        Validator("ANSIBLETOWER.results_limit", is_type_of=int, default=20),
        Validator("ANSIBLETOWER.error_scope", default="last"),
        Validator("ANSIBLETOWER.base_url", must_exist=True),
        # Validator combination for username+password or token
        (
            (
                Validator("ANSIBLETOWER.username", must_exist=True)
                & Validator("ANSIBLETOWER.password", must_exist=True)
            )
            | Validator("ANSIBLETOWER.token", must_exist=True)
        ),
        Validator("ANSIBLETOWER.inventory", default=None),
        Validator("ANSIBLETOWER.AAP_VERSION", default="", cast=str),
        Validator("ANSIBLETOWER.max_resilient_wait", is_type_of=int, default=7200),
        Validator(
            "ANSIBLETOWER.dangling_behavior",
            default="checkin",
            is_in=["prompt", "checkin", "store"],
        ),
    ]

    _checkout_options = [
        click.option(
            "--tower-inventory",
            type=str,
            help="AnsibleTower inventory to checkout a host on",
        ),
        click.option("--workflow", type=str, help="Name of a workflow used to checkout a host"),
    ]
    _execute_options = [
        click.option(
            "--tower-inventory",
            type=str,
            help="AnsibleTower inventory to execute against",
        ),
        click.option("--workflow", type=str, help="Name of a workflow to execute"),
        click.option("--job-template", type=str, help="Name of a job template to execute"),
    ]
    _extend_options = [
        click.option(
            "--new-expire-time",
            type=str,
            help="Time host should expire or time added to host reservation.",
        ),
    ]

    _sensitive_attrs = ["pword", "password", "token"]

    def __init__(self, **kwargs):
        """Almost all values are taken from Broker's config with the following exceptions.

        kwargs:
            tower_inventory: AnsibleTower inventory to use for this instance
            config: awxkit config object
            root: awxkit api root object
        """
        super().__init__(**kwargs)

        # Import awxkit on first use
        global awxkit  # noqa: PLW0603
        if awxkit is None:
            try:
                import awxkit as awxkit_module

                awxkit = awxkit_module
                # Configure ruamel yaml representer for PseudoNamespace
                yaml.representer.add_representer(
                    awxkit.utils.PseudoNamespace,
                    lambda dumper, data: dumper.represent_dict(dict(data)),
                )
            except ImportError as err:
                raise exceptions.UserError(
                    message="Unable to import awxkit. Is it installed? Install with 'pip install awxkit' or 'pip install broker[ansibletower]'"
                ) from err

        # get our instance settings
        self.url = self._settings.ANSIBLETOWER.base_url
        self.uname = self._settings.ANSIBLETOWER.get("username")
        self.pword = self._settings.ANSIBLETOWER.get("password")
        self.token = self._settings.ANSIBLETOWER.get("token")
        self.dangling_behavior = self._settings.ANSIBLETOWER.get("dangling_behavior")
        self._inventory = kwargs.get("tower_inventory") or self._settings.ANSIBLETOWER.inventory
        # Init the class itself
        config = kwargs.get("config")
        root = kwargs.get("root")
        self._v2, self._temp_token_desc = get_awxkit_and_uname(
            awxkit_config=config,
            root=root,
            url=self.url,
            token=self.token,
            uname=self.uname,
            pword=self.pword,
            broker_settings=self._settings,
        )

        # Get the username for the authenticated user
        # If a username was specified in config, use that instead
        self.username = self.uname or self._v2.me.get().results[0].username
        # Check to see if we're running AAP (ver 4.0+)
        self._is_aap = self._v2.ping.get().version[0] != "3"

    def __del__(self):
        """Clean up any temporary tokens we created."""
        if getattr(self, "_temp_token_desc", None) and hasattr(self, "_v2"):
            try:
                # Find and delete the temporary token
                tokens = self._v2.tokens.get(description=self._temp_token_desc).results
                if tokens:
                    for token in tokens:
                        token.delete()
                    logger.debug(f"Deleted temporary token: {self._temp_token_desc}")
            except Exception as err:  # noqa: BLE001 - Not currently known
                # Just log the error since we're in __del__
                logger.debug(f"Failed to delete temporary token: {err}")

    @staticmethod
    def _pull_params(kwargs):
        """Given a kwarg dict, separate AT-specific parameters from other kwargs.

        AT-specific params must stat with double underscores.
        Example: __page_size.
        """
        params, new_kwargs = {}, {}
        for key, value in kwargs.items():
            if key.startswith("__"):
                params[key[2:]] = value
            else:
                new_kwargs[key] = value
        return params, new_kwargs

    def _host_release(self):
        caller_host = inspect.stack()[1][0].f_locals["host"]
        broker_args = getattr(caller_host, "_broker_args", {}).get("_broker_args", {})
        # remove the workflow field since it will conflict with the release workflow
        broker_args.pop("workflow", None)
        # reconstruct tower provider inventory information when needed
        prov_inv = broker_args.pop("tower_inventory", None)
        if isinstance(prov_inv, str):
            logger.debug(f"prov_inv: {prov_inv}")
            prov_inv = self._translate_inventory(prov_inv)
        if prov_inv:
            logger.debug(f"prov_inv: {prov_inv}")
            broker_args["inventory"] = prov_inv
        source_vm = broker_args.pop("source_vm", caller_host.name)
        caller_host._prov_inst.release(source_vm, broker_args)

    def _set_attributes(self, host_inst, broker_args=None, misc_attrs=None):
        host_inst.__dict__.update(
            {
                "release": self._host_release,
                "_prov_inst": self,
                "_broker_provider": "AnsibleTower",
                "_broker_args": convert_pseudonamespaces(broker_args),
            }
        )
        if isinstance(misc_attrs, dict):
            host_inst.__dict__.update(convert_pseudonamespaces(misc_attrs))

    def _translate_inventory(self, inventory):
        if isinstance(inventory, int):  # already an id, silly
            if (inventory_info := self._v2.inventory.get(id=inventory)).results:
                return inventory_info.results[0].name
            else:
                raise ATInventoryError(
                    message=f"Unknown AnsibleTower inventory by id {inventory}",
                )
        elif isinstance(inventory, str):
            if inventory_info := self._v2.inventory.get(search=inventory):
                if inventory_info.count > 1:
                    # let's try to manually narrow down to one result if the api returns multiple
                    filtered = [inv for inv in inventory_info.results if inv.name == inventory]
                    if len(filtered) == 1:
                        return filtered[0].id
                    raise ATInventoryError(
                        message=f"Ambigious AnsibleTower inventory name {inventory}",
                    )
                elif inventory_info.count == 1:
                    return inventory_info.results.pop().id
                else:
                    raise ATInventoryError(
                        message=f"Unknown AnsibleTower inventory {inventory}",
                    )
        elif inv_id := getattr(inventory, "id", None):
            return inv_id
        elif inv_name := getattr(inventory, "name", None):
            return inv_name
        else:
            caller_context = inspect.stack()[1][0].f_locals
            raise ATInventoryError(
                message=f"Ambiguous AnsibleTower inventory {inventory} passed from {caller_context}",
            )

    def _merge_artifacts(self, at_object, strategy="last", artifacts=None):
        """Gather and merge all artifacts associated with an object and its children.

        :param at_object: object you want to merge

        :param strategy:
            strategies:
               - merge: merge artifact dictionaries together
               - last: return only the artifacts associated with the last child job

        :param artifacts: default to none

        :return: dictionary of merged artifact, used for constructing host
        """
        logger.debug(f"Attempting to merge: {at_object.name}")

        if artifacts is None:
            artifacts = {}

        # Merge with or overwrite previous artifacts, depending on strategy
        if getattr(at_object, "artifacts", None):
            logger.debug(f"Found artifacts: {at_object.artifacts}")
            if strategy == "merge":
                artifacts = helpers.merge_dicts(artifacts, at_object.artifacts)
            elif strategy == "last":
                artifacts = at_object.artifacts

        # If this is a workflow job, then find any children jobs
        if "workflow_nodes" in at_object.related:
            children = at_object.get_related("workflow_nodes").results

            # Filter out children with no associated job
            children = list(
                filter(lambda child: getattr(child.summary_fields, "job", None), children)
            )

            # Sort children by job id
            children.sort(key=lambda child: child.summary_fields.job.id)

            if strategy == "last":
                # Filter out all but the last job
                children = children[-1:]

            for child in children:
                if child.type == "workflow_job_node":
                    logger.debug(child)
                    child_id = child.summary_fields.job.id
                    child_obj = self._v2.jobs.get(id=child_id).results
                    if child_obj:
                        child_obj = child_obj.pop()
                        artifacts = (
                            self._merge_artifacts(child_obj, strategy, artifacts) or artifacts
                        )
                    else:
                        logger.warning(
                            f"Unable to pull information from child job with id {child_id}."
                        )
        return artifacts

    def _get_failure_messages(self, workflow):
        """Find all failure nodes and aggregate failure messages."""
        failure_messages = []
        # get all failed job nodes (iterate)
        if "workflow_nodes" in workflow.related:
            children = workflow.get_related("workflow_nodes").results
            # filter out children with no associated job
            children = list(
                filter(lambda child: getattr(child.summary_fields, "job", None), children)
            )
            # filter out children that didn't fail
            children = list(filter(lambda child: child.summary_fields.job.failed, children))
            children.sort(key=lambda child: child.summary_fields.job.id)
            for child in children[::-1]:
                if child.type == "workflow_job_node":
                    logger.debug(child)
                    child_id = child.summary_fields.job.id
                    child_obj = self._v2.jobs.get(id=child_id).results
                    if child_obj:
                        child_obj = child_obj.pop()
                        if child_obj.status == "error":
                            failure_messages.append(
                                {
                                    "job": child_obj.name,
                                    "reason": getattr(
                                        child_obj,
                                        "result_traceback",
                                        child_obj.job_explanation,
                                    ),
                                }
                            )
                        else:
                            # get all failed job_events for each job (filter failed=true)
                            failed_events = [
                                ev
                                for ev in child_obj.get_related("job_events", page_size=200).results
                                if ev.failed
                            ]
                            # find the one(s) with event_data['res']['msg']
                            failure_messages.extend(
                                [
                                    {
                                        "job": child_obj.name,
                                        "task": ev.event_data["play"],
                                        "reason": ev.event_data["res"]["msg"],
                                    }
                                    for ev in failed_events
                                    if ev.event_data.get("res", {}).get("msg")
                                ]
                            )
        if not failure_messages:
            return {
                "reason": f"Unable to determine failure cause for {workflow.name} ar {workflow.url}"
            }
        if self._settings.ANSIBLETOWER.error_scope == "last":
            return failure_messages[0]
        else:
            return failure_messages

    def _try_get_dangling_hosts(self, failed_workflow):
        """Get one or more hosts that may have been left behind by a failed workflow."""
        hosts = []
        for node in failed_workflow.get_related("workflow_nodes").results:
            if not (job_fields := node.summary_fields.get("job", {})) or job_fields.get(
                "failed"
            ):  # skip jobs with no summary fields and failed jobs
                continue
            if jobs := self._v2.jobs.get(id=job_fields["id"]).results:
                if vm_name := jobs[0].artifacts.get("vm_name"):
                    hosts.append(vm_name)
        return list(set(hosts))

    def handle_dangling_hosts(self, job):
        """Attempt to check in dangling hosts associated with the given job."""
        dangling_hosts = self._try_get_dangling_hosts(job)
        if not dangling_hosts:
            logger.debug("No dangling hosts found for the failed job.")
            return
        dangling_behavior = self.dangling_behavior
        for dangling_host in dangling_hosts:
            logger.warning(f"Found dangling host: {dangling_host}")
            if dangling_behavior == "prompt":
                choice = Prompt.ask(
                    "What would you like to do with this host? [c/s/cA/sA]\n"
                    "Checkin (c), Store (s), Checkin All (cA), Store All (sA)",
                    choices=["c", "s", "cA", "sA"],
                )
                if choice == "cA":
                    dangling_behavior = "checkin"
                elif choice == "sA":
                    dangling_behavior = "store"
            else:
                choice = None
            # handle checkins
            if choice == "c" or dangling_behavior == "checkin":
                try:
                    logger.info(f"Checking in dangling host: {dangling_host}")
                    self.release(dangling_host)
                    logger.info(f"Successfully checked in dangling host: {dangling_host}")
                except exceptions.BrokerError:
                    logger.error(f"Failed to check in dangling host: {dangling_host}")
            elif choice == "s" or dangling_behavior == "store":
                logger.info(f"Storing dangling host: {dangling_host}")
                host = self._v2.hosts.get(name=dangling_host).results[0]
                host = self._compile_host_info(host)
                host["deploy_failed"] = True
                update_inventory(add=host)

    def _compile_host_info(self, host):
        try:
            host_facts = host.related.ansible_facts.get()
        except awxkit.exceptions.Forbidden as err:
            logger.warning(f"Unable to get facts for {host.name}: {err}")
            host_facts = {}

        # Get the hostname from host variables or facts
        hostname = (
            host.variables.get("fqdn")
            or getattr(host_facts, "ansible_fqdn", None)
            # Workaround for OSP hosts that have lost their hostname
            or host.variables.get("openstack", {}).get("metadata", {}).get("fqdn", None)
        )

        # Get broker_args from host facts if present
        broker_args = getattr(host_facts, "_broker_args", {})
        broker_facts = getattr(host_facts, "_broker_facts", {})

        host_info = {key: val for key, val in broker_facts.items() if val}
        host_info.update(
            {
                "name": host.name,
                "type": host.type,
                "hostname": hostname,
                "ip": host.variables.get("ansible_host"),
                "tower_inventory": self._translate_inventory(host.inventory),
                "_broker_provider": "AnsibleTower",
                "_broker_provider_instance": self.instance,
                # Get _broker_args from host facts if present
                "_broker_args": {key: val for key, val in broker_args.items() if val},
            }
        )

        # Find and add extra fields
        interfaces = getattr(host_facts, "ansible_interfaces", [])
        facts = {
            "os_distribution": getattr(host_facts, "ansible_distribution", None),
            "os_distribution_version": getattr(host_facts, "ansible_distribution_version", None),
            "reported_devices": {"nics": interfaces} if interfaces else None,
        }
        host_info.update({key: val for key, val in facts.items() if val})

        return host_info

    @staticmethod
    def _pull_extra_vars(extra_vars):
        """Pull extra vars from a json string or pseudo-dictionary."""
        if not extra_vars:
            return {}
        try:
            return json.loads(extra_vars)
        except json.JSONDecodeError:
            logger.warning(
                f"Job uses non-json extra_vars:\n{extra_vars}\n"
                "Attempting to parse as pseudo-dictionary."
            )
            compiled = {}
            for line in extra_vars.splitlines():
                key, val = line.split(": ")
                compiled[key] = val
            return compiled

    def _resolve_labels(self, labels, target):
        """Fetch and return ids of the given labels.

        If label does not exist, create it under the same org as the target template.
        """
        label_ids = []
        for label in labels:
            label_expanded = f"{label}={labels[label]}" if labels[label] else label
            try:
                result = self._v2.labels.post(
                    {"name": label_expanded, "organization": target.summary_fields.organization.id}
                )
                if result:
                    label_ids.append(result.id)
            except awxkit.exceptions.Duplicate:
                logger.debug(f"Provider label {label_expanded} already exists on AAP instance")
                if result := self._v2.labels.get(name=label_expanded).results:
                    logger.debug(f"Provider label {label_expanded} retrieved successfully")
                    label_ids.append(result[0].id)
                else:
                    logger.warning(
                        f"Provider label {label_expanded} not found on AAP despite AAP returning 400: Duplicate while trying to create it"
                    )
        return label_ids

    def _verify_default_inventory(self, inv_id, subject):
        """Verify that the user at least has view access to a workflow/template's default inventory."""
        if inv_id is None:
            raise exceptions.UserError(
                f"The {subject} has no default inventory set, please specify one."
            )
        try:
            self._translate_inventory(inv_id)
        except ATInventoryError as err:
            raise exceptions.UserError(
                f"You don't have access to this {subject}'s default inventory, please specify one."
            ) from err

    @cached_property
    def inventory(self):
        """Return the current tower inventory."""
        if not self._inventory:
            return
        elif isinstance(self._inventory, int):
            # inventory already resolved as id
            return self._inventory
        self._inventory = self._translate_inventory(self._inventory)
        return self._inventory

    def construct_host(self, provider_params, host_classes, **kwargs):
        """Construct a host to be read by Ansible Tower.

        :param provider_params: dictionary of what the provider returns when initially
        creating the vm

        :param host_classes: host object

        :return: broker object of constructed host instance
        """
        broker_args = kwargs.copy()
        broker_facts = {}

        strategy = broker_args.pop("strategy", "last")

        def _get_fields_from_facts(facts):
            hostname = None
            name = None
            host_type = "host"

            for key, value in facts.items():
                if key.endswith("fqdn") and not hostname:
                    hostname = value if not isinstance(value, list) else value[0]
                if key in ("name", "vm_provisioned") and not name:
                    name = value if not isinstance(value, list) else value[0]
                if key.endswith("host_type"):
                    host_type = value if value in host_classes else host_type

            return hostname, name, host_type

        if provider_params:
            job = provider_params
            artifacts = self._merge_artifacts(job, strategy=strategy)

            # Get host facts from job artifacts
            if "_broker_args" in artifacts and "_broker_facts" in artifacts:
                broker_args = {k: v for k, v in artifacts._broker_args.items() if v}
                broker_facts = {k: v for k, v in artifacts._broker_facts.items() if v}
                logger.debug(artifacts)

                # Get hostname, VM name, and host type
                hostname, name, host_type = _get_fields_from_facts(broker_facts)
                if not hostname:
                    logger.warning(f"No hostname found in job artifacts:\n{artifacts}")
                logger.debug(f"hostname: {hostname}, name: {name}, host type: {host_type}")

                host_inst = host_classes[host_type](
                    hostname=hostname, name=name, broker_settings=self._settings, **broker_args
                )
                broker_facts["name"] = name
                broker_facts["hostname"] = hostname
            else:
                logger.debug(f"Host facts not found in artifacts for job: {job}")

        else:
            host_inst = host_classes[kwargs.get("type")](**broker_args)

        self._set_attributes(host_inst, broker_args=broker_args, misc_attrs=broker_facts)

        return host_inst

    @Provider.register_action("workflow", "job_template")
    def execute(self, **kwargs):  # noqa: PLR0912,PLR0915 - Possible TODO refactor
        """Execute workflow or job template in Ansible Tower.

        :param kwargs: workflow or job template name passed in a string

        :return: dictionary containing all information about executed workflow/job template
        """
        if name := kwargs.get("workflow"):
            subject = "workflow"
            get_path = self._v2.workflow_job_templates
            origin = find_origin()
            kwargs["_broker_origin"] = origin[0]
            if origin[1]:
                kwargs["_jenkins_url"] = origin[1]
        elif name := kwargs.get("job_template"):
            subject = "job_template"
            get_path = self._v2.job_templates
        else:
            raise exceptions.UserError(message="No workflow or job template specified")
        try:
            candidates = get_path.get(name=name).results
        except awxkit.exceptions.Unauthorized as err:
            raise exceptions.AuthenticationError(err.args[0]) from err
        if candidates:
            target = candidates.pop()
        else:
            raise exceptions.UserError(message=f"{subject.capitalize()} not found by name: {name}")
        payload = {}
        if inventory := kwargs.pop("inventory", None):
            payload["inventory"] = inventory
            logger.info(f"Using tower inventory: {self._translate_inventory(inventory)}")

        elif self.inventory:
            payload["inventory"] = self.inventory
            logger.info(f"Using tower inventory: {self._translate_inventory(self.inventory)}")
        else:
            self._verify_default_inventory(inv_id=getattr(target, "inventory", None), subject=name)
            logger.info("No inventory specified, Ansible Tower will use a default.")

        # provider labels handling

        provider_labels = kwargs.get("provider_labels", {})
        # include eventual common labels, specified at each level of configuration
        # typically imported from dynaconf env vars
        provider_labels.update(self._settings.get("provider_labels", {}))
        provider_labels.update(self._settings.ANSIBLETOWER.get("provider_labels", {}))
        if provider_labels:
            payload["labels"] = self._resolve_labels(provider_labels, target)
            kwargs["provider_labels"] = provider_labels

        # Save custom, non-workflow extra vars to a named variable.
        # The workflow can save these values to job artifacts / host facts.
        workflow_extra_vars = self._pull_extra_vars(target.extra_vars)
        kwargs["_broker_extra_vars"] = {
            k: v for k, v in kwargs.items() if k not in workflow_extra_vars
        }
        payload["extra_vars"] = json.dumps(kwargs)
        logger.debug(
            f"Launching {subject}: {url_parser.urljoin(self.url, str(target.url))}\n{payload=}"
        )
        job = target.launch(payload=payload)
        job_number = job.url.rstrip("/").split("/")[-1]
        job_api_url = url_parser.urljoin(self.url, str(job.url))
        # Need to change the url subject for job templates. If this increases, then come up with a better solution
        job_ui_url = url_parser.urljoin(
            self.url,
            f"/{AAP_URL_PREFIX}/jobs/{'playbook' if subject == 'job_template' else subject}/{job_number}/output",
        )
        helpers.emit(api_url=job_api_url, ui_url=job_ui_url)
        logger.info(f"Waiting for job: \nAPI: {job_api_url}\nUI: {job_ui_url}")
        resilient_job_wait(job)
        if job.status != "successful":
            failure_message = self._get_failure_messages(job)
            message_data = {
                f"{subject.capitalize()} Status": job.status,
                "Reason(s)": failure_message,
                "URL": job_ui_url,
            }
            helpers.emit(message_data)
            # handle potential dangling hosts
            if not isinstance(failure_message, list):
                failure_message = [failure_message]
            if not any("was automatically checked-in" in msg["reason"] for msg in failure_message):
                self.handle_dangling_hosts(job)
            else:
                logger.warning(f"Apparently it is in the failure message...\n{failure_message}")
            raise JobExecutionError(message_data=message_data["Reason(s)"])
        if strategy := kwargs.pop("artifacts", None):
            return self._merge_artifacts(job, strategy=strategy)
        return job

    def get_inventory(self, user=None):
        """Compile a list of hosts based on any inventory a user's name is mentioned."""
        user = user or self.uname
        invs = [
            inv
            for inv in self._v2.inventory.get(page_size=200).results
            if user in inv.name or user == "@ll"
        ]
        hosts = []
        for inv in invs:
            inv_hosts = inv.get_related("hosts", page_size=200).results
            hosts.extend(inv_hosts)
        with click.progressbar(hosts, label="Compiling host information") as hosts_bar:
            compiled_host_info = [self._compile_host_info(host) for host in hosts_bar]
        return compiled_host_info

    def extend(self, target_vm, new_expire_time=None, provider_labels=None):
        """Run the extend workflow with defaults args.

        :param target_vm: This should be a host object
        """
        if provider_labels is None:
            provider_labels = {}
        # check if an inventory was specified. if so overwrite the current inventory
        if new_inv := target_vm._broker_args.get("tower_inventory"):
            if new_inv != self._inventory:
                self._inventory = new_inv
                if hasattr(self.__dict__, "inventory"):
                    del self.inventory  # clear the cached value
        return self.execute(
            workflow=self._settings.ANSIBLETOWER.extend_workflow,
            target_vm=target_vm.name,
            new_expire_time=new_expire_time or self._settings.ANSIBLETOWER.get("new_expire_time"),
            provider_labels=provider_labels,
        )

    @Provider.help_override(tower_inventory="Filter results by the specified inventory.")
    def provider_help(  # noqa: PLR0911, PLR0912, PLR0915 - Possible TODO refactor
        self,
        workflows=False,
        workflow=None,
        job_templates=False,
        job_template=None,
        templates=False,
        inventories=False,
        inventory=None,
        flavors=False,
        tower_inventory=None,
        **kwargs,
    ):
        """Get a list of extra vars and their defaults from a workflow."""
        results_limit = kwargs.get("results_limit", self._settings.ANSIBLETOWER.results_limit)
        rich_console = Console(no_color=self._settings.less_colors)
        # if a user passes a different tower inventory, let's try using that
        self._inventory = tower_inventory
        if workflow:
            if wfjt := self._v2.workflow_job_templates.get(name=workflow).results:
                wfjt = wfjt.pop()
            else:
                logger.warning(f"Workflow {workflow} not found!")
                return
            default_inv = self._v2.inventory.get(id=wfjt.inventory).results.pop()
            top_table = helpers.dict_to_table(
                {"Description": wfjt.description, "Inventory": default_inv["name"]},
                title=f"{workflow} information",
            )
            rich_console.print(top_table)
            extras_table = helpers.dict_to_table(
                json.loads(wfjt.extra_vars),
                title="Workflow Variables",
                headers=("Variable", "Default Value"),
            )
            rich_console.print(extras_table)
        elif workflows:
            workflows = [
                workflow.name
                for workflow in self._v2.workflow_job_templates.get(page_size=1000).results
                if workflow.summary_fields.user_capabilities.get("start")
            ]
            if not workflows:
                logger.warning("No workflows found")
                return
            if res_filter := kwargs.get("results_filter"):
                workflows = eval_filter(workflows, res_filter, "res")
                workflows = workflows if isinstance(workflows, list) else [workflows]
            workflow_table = helpers.dictlist_to_table(
                [{"name": workflow} for workflow in workflows[:results_limit]],
                title="Available Workflows",
                _id=False,
                headers=False,
            )
            rich_console.print(workflow_table)
        elif inventory:
            if inv := self._v2.inventory.get(name=inventory, kind="").results:
                inv = inv.pop()
            else:
                logger.warning(f"Inventory {inventory} not found!")
                return
            inv_table = helpers.dict_to_table(
                {"ID": inv.id, "Name": inv.name, "Description": inv.description},
                title="Inventory Details",
            )
            rich_console.print(inv_table)
        elif inventories:
            inv = [inv.name for inv in self._v2.inventory.get(kind="", page_size=1000).results]
            if not inv:
                logger.warning("No inventories found!")
                return
            if res_filter := kwargs.get("results_filter"):
                inv = eval_filter(inv, res_filter, "res")
                inv = inv if isinstance(inv, list) else [inv]
            inv_table = helpers.dictlist_to_table(
                [{"name": i} for i in inv[:results_limit]],
                title="Available Inventories",
                _id=False,
                headers=False,
            )
            rich_console.print(inv_table)
        elif job_template:
            if jt := self._v2.job_templates.get(name=job_template).results:
                jt = jt.pop()
            else:
                logger.warning(f"Job Template {job_template} not found!")
                return
            default_inv = self._v2.inventory.get(id=jt.inventory).results.pop()
            top_table = helpers.dict_to_table(
                {"Description": jt.description, "Inventory": default_inv["name"]},
                title=f"{job_template} information",
            )
            rich_console.print(top_table)
            extras_table = helpers.dict_to_table(
                json.loads(jt.extra_vars),
                title="Job Template Variables",
                headers=("Variable", "Default Value"),
            )
            rich_console.print(extras_table)
        elif job_templates:
            job_templates = [
                job_template.name
                for job_template in self._v2.job_templates.get(page_size=1000).results
                if job_template.summary_fields.user_capabilities.get("start")
            ]
            if not job_templates:
                logger.warning("No job templates found!")
                return
            if res_filter := kwargs.get("results_filter"):
                job_templates = eval_filter(job_templates, res_filter, "res")
                job_templates = (
                    job_templates if isinstance(job_templates, list) else [job_templates]
                )
            job_template_table = helpers.dictlist_to_table(
                [{"name": job_template} for job_template in job_templates[:results_limit]],
                title="Available Job Templates",
                _id=False,
                headers=False,
            )
            rich_console.print(job_template_table)
        elif templates:
            templates = list(
                set(
                    self.execute(workflow="list-templates", artifacts="last")["data_out"][
                        "list_templates"
                    ]
                )
            )
            if not templates:
                logger.warning("No templates found!")
                return
            templates.sort(reverse=True)
            if res_filter := kwargs.get("results_filter"):
                templates = eval_filter(templates, res_filter, "res")
                templates = templates if isinstance(templates, list) else [templates]
            template_table = helpers.dictlist_to_table(
                [{"name": template} for template in templates[:results_limit]],
                title="Available Templates",
                _id=False,
                headers=False,
            )
            rich_console.print(template_table)
        elif flavors:
            flavors = self.execute(workflow="list-flavors", artifacts="last")["data_out"][
                "list_flavors"
            ]
            if not flavors:
                logger.warning("No flavors found!")
                return
            if res_filter := kwargs.get("results_filter"):
                flavors = eval_filter(flavors, res_filter, "res")
                flavors = flavors if isinstance(flavors, list) else [flavors]
            flavor_table = helpers.dictlist_to_table(
                flavors[:results_limit], title="Available Flavors", _id=False
            )
            rich_console.print(flavor_table)

    def release(self, name, broker_args=None):
        """Release the host back to the tower instance via the release workflow."""
        if broker_args is None:
            broker_args = {}
        return self.execute(
            workflow=self._settings.ANSIBLETOWER.release_workflow,
            source_vm=name,
            **broker_args,
        )



"""Module providing base SSH methods and classes."""

import socket

from logzero import logger

from broker import exceptions


def _create_connect_socket(host, port, timeout, ipv6=False, ipv4_fallback=True, sock=None):
    """Create a socket and establish a connection to the specified host and port.

    Args:
        host (str): The hostname or IP address of the remote server.
        port (int): The port number to connect to.
        timeout (float): The timeout value in seconds for the socket connection.
        ipv6 (bool, optional): Whether to use IPv6. Defaults to False.
        ipv4_fallback (bool, optional): Whether to fallback to IPv4 if IPv6 fails. Defaults to True.
        sock (socket.socket, optional): An existing socket object to use. Defaults to None.

    Returns:
        socket.socket: The connected socket object.
        bool: True if IPv6 was used, False otherwise.

    Raises:
        exceptions.ConnectionError: If unable to establish a connection to the host.
    """
    if ipv6 and not sock:
        try:
            sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        except OSError as err:
            if ipv4_fallback:
                logger.warning(f"IPv6 failed with {err}. Falling back to IPv4.")
                return _create_connect_socket(host, port, timeout, ipv6=False)
            else:
                raise exceptions.ConnectionError(
                    f"Unable to establish IPv6 connection to {host}."
                ) from err
    elif not sock:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    if ipv6:
        try:
            sock.connect((host, port))
        except socket.gaierror as err:
            if ipv4_fallback:
                logger.warning(f"IPv6 connection failed to {host}. Falling back to IPv4.")
                # FIXME this socket was created for AF_INET6. We shouldn't reuse it with ipv6=False.
                return _create_connect_socket(host, port, timeout, ipv6=False, sock=sock)
            else:
                raise exceptions.ConnectionError(
                    f"Unable to establish IPv6 connection to {host}."
                ) from err
    else:
        sock.connect((host, port))
    return sock, ipv6



"""A collection of Broker-specific exceptions."""

import logging

from logzero import logger


class BrokerError(Exception):
    """Base class for Broker exceptions."""

    error_code = 1

    def __init__(self, message="An unhandled exception occured!"):
        # Log the exception if the logger is set to DEBUG
        if logger.level == logging.DEBUG and isinstance(message, Exception):
            logger.exception(message)
        self.message = message
        logger.error(f"{self.__class__.__name__}: {self.message}")


class AuthenticationError(BrokerError):
    """Raised when authentication with a provider or Host fails."""

    error_code = 5


class PermissionError(BrokerError):
    """Raised when the user does not have permission to perform an action."""

    error_code = 6


class ProviderError(BrokerError):
    """Raised when a provider-specific error occurs."""

    error_code = 7

    def __init__(self, provider=None, message="Unspecified exception"):
        self.message = f"{provider} encountered the following error: {message}"
        super().__init__(message=self.message)


class ConfigurationError(BrokerError):
    """Raised when a Broker configuration error occurs."""

    error_code = 8


class NotImplementedError(BrokerError):
    """Raised when a method or function has not been implemented."""

    error_code = 9


class HostError(BrokerError):
    """Raised when a Host-specific error occurs."""

    error_code = 10

    def __init__(self, host=None, message="Unspecified exception"):
        if host:
            self.message = f"{host.hostname or host.name}: {message}"
        super().__init__(message=self.message)


class ContainerBindError(BrokerError):
    """Raised when a problem occurs at the container's bind level."""

    error_code = 11


class BeakerBindError(BrokerError):
    """Raised when a problem occurs at the Beaker bind level."""

    error_code = 12


class UserError(BrokerError):
    """Raised when a user causes an otherwise unclassified error."""

    error_code = 13


class ForemanBindError(BrokerError):
    """Raised when a problem occurs at the Foreman bind level."""

    error_code = 14


class ParamikoBindError(BrokerError):
    """Raised when a problem occurs at the Paramiko bind level."""

    error_code = 15



"""Foreman provider implementation."""

import time

from logzero import logger
import requests

from broker import exceptions
from broker.settings import clone_global_settings


class ForemanBind:
    """Default runtime to query Foreman."""

    headers = {
        "Content-Type": "application/json",
    }

    def __init__(self, broker_settings=None, **kwargs):
        self._settings = broker_settings or clone_global_settings()
        self.foreman_username = kwargs.get(
            "foreman_username", self._settings.foreman.foreman_username
        )
        self.foreman_password = kwargs.get(
            "foreman_password", self._settings.foreman.foreman_password
        )
        self.url = kwargs.get("url", self._settings.foreman.foreman_url)
        self.prefix = kwargs.get("prefix", self._settings.foreman.name_prefix)
        self.verify = kwargs.get("verify", self._settings.foreman.verify)

        self.session = requests.session()

    def _interpret_response(self, response):
        """Handle responses from Foreman, in particular catch errors."""
        if "error" in response:
            error = response["error"]
            message = error.get("message")
            if message is not None and "Unable to authenticate user" in message:
                raise exceptions.AuthenticationError(message)
            raise exceptions.ForemanBindError(
                " ".join(error["full_messages"]),
            )
        if "errors" in response:
            raise exceptions.ForemanBindError(" ".join(response["errors"]["base"]))
        return response

    def _get(self, endpoint):
        """Send GET request to Foreman API."""
        response = self.session.get(
            self.url + endpoint,
            auth=(self.foreman_username, self.foreman_password),
            headers=self.headers,
            verify=self.verify,
        ).json()
        return self._interpret_response(response)

    def _post(self, endpoint, **kwargs):
        """Send POST request to Foreman API."""
        response = self.session.post(
            self.url + endpoint,
            auth=(self.foreman_username, self.foreman_password),
            headers=self.headers,
            verify=self.verify,
            **kwargs,
        ).json()
        return self._interpret_response(response)

    def _delete(self, endpoint, **kwargs):
        """Send DELETE request to Foreman API."""
        response = self.session.delete(
            self.url + endpoint,
            auth=(self.foreman_username, self.foreman_password),
            headers=self.headers,
            verify=self.verify,
            **kwargs,
        )
        return self._interpret_response(response)

    def obtain_id_from_name(self, resource_type, resource_name):
        """Obtain id for resource with given name.

        :param resource_type: Resource type, like hostgroups, hosts, ...

        :param resource_name: String-like identifier of the resource

        :return: ID of the found object
        """
        response = self._get(
            f"/api/{resource_type}?per_page=200",
        )
        try:
            result = response["results"]
            resource = next(
                x
                for x in result
                if x.get("title") == resource_name or x.get("name") == resource_name
            )
            id_ = resource["id"]
        except KeyError:
            logger.error(f"Could not find {resource_type} {resource_name}")
            raise
        except StopIteration:
            raise exceptions.ForemanBindError(
                f"Could not find {resource_name} in {resource_type}",
            )
        return id_

    def create_job_invocation(self, data):
        """Run a job from the provided data."""
        return self._post(
            "/api/job_invocations",
            json=data,
        )["id"]

    def job_output(self, job_id):
        """Return output of job."""
        return self._get(f"/api/job_invocations/{job_id}/outputs")["outputs"][0]["output"]

    def wait_for_job_to_finish(self, job_id):
        """Poll API for job status until it is finished.

        :param job_id: id of the job to poll
        """
        still_running = True
        while still_running:
            response = self._get(f"/api/job_invocations/{job_id}")
            still_running = response["status_label"] == "running"
            time.sleep(1)

    def hostgroups(self):
        """Return list of available hostgroups."""
        return self._get("/api/hostgroups")

    def hostgroup(self, name):
        """Return list of available hostgroups."""
        hostgroup_id = self.obtain_id_from_name("hostgroups", name)
        return self._get(f"/api/hostgroups/{hostgroup_id}")

    def hosts(self):
        """Return list of hosts deployed using this prefix."""
        return self._get(f"/api/hosts?search={self.prefix}")["results"]

    def image_uuid(self, compute_resource_id, image_name):
        """Return the uuid of a VM image on a specific compute resource."""
        try:
            return self._get(
                f"/api/compute_resources/{compute_resource_id}/images/?search=name={image_name}"
            )["results"][0]["uuid"]
        except IndexError:
            raise exceptions.ForemanBindError(f"Could not find {image_name} in VM images")

    def create_host(self, data):
        """Create a host from the provided data."""
        return self._post("/api/hosts", json=data)

    def wait_for_host_to_install(self, hostname):
        """Poll API for host build status until it is built.

        :param hostname: name of the host which is currently being built
        """
        building = True
        while building:
            host_status = self._get(f"/api/hosts/{hostname}")
            building = host_status["build_status"] != 0
            time.sleep(1)



"""Module for managing hosts.

This module defines the `Host` class, which represents a host that can be accessed via SSH or Bind.
The `Host` class provides methods for connecting to the host, executing commands, and transferring files.
It additionally exposes a common interface for Broker to manage host creation, checkin, and deletion.
It is recommended to subclass the `Host` class for custom behavior.

Usage:
    To use the `Host` class, create a new `Host` object with the required parameters:

    ```
    from broker.hosts import Host

    host = Host(hostname="example.com", username="user", password="password")
    ```
"""

from logzero import logger

from broker.exceptions import HostError, NotImplementedError

SETTINGS_VALIDATED = False


class Host:
    """Class representing a host that can be accessed via SSH or Bind.

    This class provides methods for connecting to the host, executing commands, and transferring files.
    It additionally exposes a common interface for Broker to manage host creation, checkin, and deletion.
    It is recommended to subclass the `Host` class for custom behavior.
    """

    DEFAULT_TIMEOUT = 0  # timeout in ms, 0 is infinite
    DEFAULT_USER = "root"
    keep_keys = (
        "hostname",
        "_broker_provider",
        "_broker_args",
        "tower_inventory",
        "deploy_network_type",
        "job_id",
        "_attrs",
        "ip",
        "os_distribution",
        "os_distribution_version",
        "reported_devices",
        "exposed_ports",
    )

    def __init__(self, broker_settings=None, **kwargs):
        """Create a Host instance.

        Expected kwargs:
            hostname: (str) - Hostname or IP address of the host, required
            name: (str) - Name of the host
            username: (str) - Username to use for SSH connection
            password: (str) - Password to use for SSH connection
            connection_timeout: (int) - Timeout for SSH connection
            port: (int) - Port to use for SSH connection
            key_filename: (str) - Path to SSH key file to use for SSH connection
            ipv6 (bool): Whether or not to use IPv6. Defaults to False.
            ipv4_fallback (bool): Whether or not to fallback to IPv4 if IPv6 fails. Defaults to True.

        If `broker_settings` is provided, it will be used over Broker's global settings.
        """
        if broker_settings:
            # Use the provided settings object instead of the global one
            self._settings = broker_settings
            logger.debug("Validating local ssh settings")
            self._settings.validators.validate(only="SSH")
        else:
            # Use the global settings object
            from broker.settings import clone_global_settings

            self._settings = clone_global_settings()
            global SETTINGS_VALIDATED  # noqa: PLW0603
            if not SETTINGS_VALIDATED:
                logger.debug("Validating ssh settings")
                self._settings.validators.validate(only="SSH")
                SETTINGS_VALIDATED = True
        logger.debug(f"Constructing host using {kwargs=}")
        self.hostname = kwargs.get("hostname") or kwargs.get("ip")
        if not self.hostname:
            # check to see if we're being reconstructued, likely for checkin
            import inspect

            if any(f.function == "reconstruct_host" for f in inspect.stack()):
                logger.debug("Ignoring missing hostname and ip for checkin reconstruction.")
            else:
                raise HostError("Host must be constructed with a hostname or ip")
        self.name = kwargs.pop("name", None)
        self.username = kwargs.pop("username", self._settings.SSH.HOST_USERNAME)
        self.password = kwargs.pop("password", self._settings.SSH.HOST_PASSWORD)
        self.timeout = kwargs.pop("connection_timeout", self._settings.SSH.HOST_CONNECTION_TIMEOUT)
        self.port = kwargs.pop("port", self._settings.SSH.HOST_SSH_PORT)
        self.key_filename = kwargs.pop("key_filename", self._settings.SSH.HOST_SSH_KEY_FILENAME)
        self.ipv6 = kwargs.pop("ipv6", self._settings.SSH.HOST_IPV6)
        self.ipv4_fallback = kwargs.pop("ipv4_fallback", self._settings.SSH.HOST_IPV4_FALLBACK)
        self.__dict__.update(kwargs)  # Make every other kwarg an attribute
        self._session = None

    def __del__(self):
        """Try to close the connection on garbage collection of the host instance."""
        if hasattr(self, "_session") and self._session is not None:
            self.close()
        # If host inherits from a different class with __del__, it should get called through super

    @property
    def session(self):
        """Return the session object for the host.

        If the session object does not exist, it will be created by calling the `connect` method.
        If the host is a non-SSH-enabled container host, a `ContainerSession` object will be created instead.
        """
        if self._session is None:
            # Check to see if we're a non-ssh-enabled Container Host
            if hasattr(self, "_cont_inst") and not self._cont_inst.ports.get(22):
                from broker.session import ContainerSession

                runtime = "podman" if "podman" in str(self._cont_inst.client) else "docker"
                self._session = ContainerSession(
                    self, runtime=runtime, broker_settings=self._settings
                )
            else:
                self.connect()
        return self._session

    def connect(self, **kwargs):
        """Create one or more connections to the target host."""
        for key, val in kwargs.items():
            setattr(self, key, val)
        if getattr(self, "is_container", None) or getattr(self, "_cont_inst", None):
            from broker.session import ContainerSession

            self._session = ContainerSession(
                self, kwargs.get("runtime"), broker_settings=self._settings
            )
        else:
            logger.debug(f"Creating SSH session to {self.hostname}")
            # Create a session using the entry-points based approach
            from broker.session import make_session

            self._session = make_session(
                broker_settings=self._settings,
                hostname=self.hostname,
                port=getattr(self, "port", self._settings.get("SSH", {}).get("HOST_SSH_PORT", 22)),
                username=getattr(
                    self,
                    "username",
                    self._settings.get("SSH", {}).get("HOST_USERNAME", self.DEFAULT_USER),
                ),
                password=getattr(
                    self,
                    "password",
                    self._settings.get("SSH", {}).get("HOST_PASSWORD", None),
                ),
                key_filename=getattr(
                    self,
                    "key_filename",
                    self._settings.get("SSH", {}).get("HOST_SSH_KEY_FILENAME", None),
                ),
                timeout=getattr(
                    self,
                    "timeout",
                    self._settings.get("SSH", {}).get("HOST_CONNECTION_TIMEOUT", 60),
                ),
                host=self.hostname,  # For hussh backend compatibility
            )

    def close(self):
        """Close the SSH connection to the host."""
        if self._session is not None:
            self._session.disconnect()
        self._session = None

    def release(self):
        """Release the host using the appropriate method for the provider."""
        raise NotImplementedError("release has not been implemented for this provider")

    # @cached_property
    def _pkg_mgr(self):
        for mgr in ["yum", "dnf", "zypper"]:
            if f"no {mgr} in" in self.execute(f"which {mgr}"):
                return mgr
        return None

    def execute(self, command, timeout=None):
        """Execute a command on the host using SSH.

        Args:
            command (str): The command to execute on the host.
            timeout (int): The timeout for the SSH connection in seconds. Defaults to `None`.

        Returns:
            str: The output of the command executed on the host.
        """
        timeout = self.DEFAULT_TIMEOUT if timeout is None else timeout
        logger.debug(f"{self.hostname} executing command: {command}")
        res = self.session.run(command, timeout=timeout)
        logger.debug(f"{self.hostname} command result:\n{res}")
        return res

    def to_dict(self):
        """Convert the host instance to a dictionary representation.

        Includes attributes specified in the `keep_keys` class attribute.
        """
        ret_dict = {
            "name": getattr(self, "name", None),
            "_broker_provider_instance": self._prov_inst.instance,
            "type": "host",
        }
        ret_dict.update({k: v for k, v in self.__dict__.items() if k in self.keep_keys})
        return ret_dict

    def setup(self):
        """Automatically ran when entering a Broker context manager."""
        pass

    def teardown(self):
        """Automatically ran when exiting a Broker context manager."""
        pass

    def __repr__(self):
        """Return a string representation of the host."""
        inner = ", ".join(
            f"{k}={v}"
            for k, v in self.__dict__.items()
            if not k.startswith("_") and not callable(v)
        )
        return f"{self.__class__.__name__}({inner})"

    @classmethod
    def from_dict(cls, arg_dict):
        """Create a Host instance from a dict."""
        return cls(**arg_dict, from_dict=True)



"""Module providing classes to establish ssh or ssh-like connections to hosts.

Classes:
    Session - Wrapper around hussh's auth/connection system.

Note: You typically want to use a Host object instance to create sessions,
      not these classes directly.
"""

from contextlib import contextmanager
from pathlib import Path

from hussh import Connection

from broker import exceptions, helpers


class Session:
    """Wrapper around hussh's auth/connection system."""

    def __init__(self, **kwargs):
        """Initialize a Session object.

        kwargs:
            hostname (str): The hostname or IP address of the remote host. Defaults to 'localhost'.
            username (str): The username to authenticate with. Defaults to 'root'.
            timeout (float): The timeout for the connection in seconds. Defaults to 60.
            port (int): The port number to connect to. Defaults to 22.
            key_filename (str): The path to the private key file to use for authentication.
            password (str): The password to use for authentication.
            ipv6 (bool): Whether or not to use IPv6. Defaults to False.
            ipv4_fallback (bool): Whether or not to fallback to IPv4 if IPv6 fails. Defaults to True.

        Raises:
            AuthException: If no password or key file is provided.
            ConnectionError: If the connection fails.
            FileNotFoundError: If the key file is not found.
        """
        host = kwargs.get("hostname", "localhost")
        user = kwargs.get("username", "root")
        port = kwargs.get("port", 22)
        timeout = kwargs.get("timeout", 60) * 1000

        key_filename = kwargs.get("key_filename")
        password = kwargs.get("password")

        # TODO Create and use socket if hussh allows user to specify one
        self.session = None

        conn_kwargs = {"username": user, "port": port, "timeout": timeout}
        try:
            if key_filename:
                auth_type = "Key"
                if not Path(key_filename).exists():
                    raise FileNotFoundError(f"Key not found in '{key_filename}'")
                conn_kwargs["private_key"] = key_filename
            elif password:
                auth_type = "Password"
                conn_kwargs["password"] = password
            elif user:
                auth_type = "Session"
            else:
                raise exceptions.AuthenticationError("No password or key file provided.")

            self.session = Connection(host, **conn_kwargs)

        except Exception as err:
            raise exceptions.AuthenticationError(
                f"{auth_type}-based authentication failed."
            ) from err

    @staticmethod
    def _set_destination(source, destination):
        dest = destination or source
        if dest.endswith("/"):
            dest = dest + Path(source).name
        return dest

    def disconnect(self):
        """Disconnect session."""

    def remote_copy(self, source, dest_host, dest_path=None, ensure_dir=True):
        """Copy a file from this host to another."""
        dest_path = dest_path or source
        if ensure_dir:
            dest_host.session.run(f"mkdir -p {Path(dest_path).absolute().parent}")

        # Copy from this host to destination host
        self.session.remote_copy(
            source_path=str(source), dest_conn=dest_host.session.session, dest_path=str(dest_path)
        )

    def run(self, command, timeout=0):
        """Run a command on the host and return the results."""
        result = self.session.execute(command, timeout=helpers.translate_timeout(timeout))
        # Create broker Result from hussh SSHResult
        return helpers.Result(
            status=result.status,
            stderr=result.stderr,
            stdout=result.stdout,
        )

    def scp_read(self, source, destination=None, return_data=False):
        """SCP read a remote file into a local destination or return a bytes object if return_data is True."""
        destination = self._set_destination(str(source), str(destination))
        if return_data:
            return self.session.scp_read(remote_path=str(source))
        self.session.scp_read(remote_path=str(source), local_path=str(destination))

    def scp_write(self, source, destination=None, ensure_dir=True):
        """SCP write a local file to a remote destination."""
        destination = self._set_destination(str(source), str(destination))
        if ensure_dir:
            self.run(f"mkdir -p {Path(destination).absolute().parent}")
        self.session.scp_write(str(source), str(destination))

    def sftp_read(self, source, destination=None, return_data=False):
        """Read a remote file into a local destination or return a bytes object if return_data is True."""
        if return_data:
            return self.session.sftp_read(remote_path=str(source)).encode("utf-8")

        destination = self._set_destination(str(source), str(destination))

        # Create the destination path if it doesn't exist
        Path(destination).parent.mkdir(parents=True, exist_ok=True)

        self.session.sftp_read(remote_path=str(source), local_path=str(destination))

    def sftp_write(self, source, destination=None, ensure_dir=True):
        """Sftp write a local file to a remote destination."""
        destination = self._set_destination(str(source), str(destination))
        if ensure_dir:
            self.run(f"mkdir -p {Path(destination).absolute().parent}")
        self.session.sftp_write(local_path=str(source), remote_path=str(destination))

    def shell(self, pty=False):
        """Create and return an interactive shell instance."""
        return self.session.shell(pty=pty)

    @contextmanager
    def tail_file(self, filename):
        """Tail a file on the remote host."""
        with self.session.tail(str(filename)) as _tailer:
            yield (tailer := FileTailer(tailer=_tailer))
        tailer.contents = _tailer.contents


class FileTailer:
    """Wrapper for hussh's FileTailer class."""

    def __init__(self, **kwargs):
        self.tailer = kwargs.get("tailer")



"""Config migrations for versions older than 0.6.3 to 0.6.3."""

from logzero import logger

TO_VERSION = "0.6.3"


def add_dangling_behavior(config_dict):
    """Add the dangling_behavior config to AnsibleTower."""
    if "AnsibleTower" in config_dict:
        if "dangling_behavior" not in config_dict["AnsibleTower"]:
            logger.debug("Adding dangling_behavior to AnsibleTower.")
            config_dict["AnsibleTower"]["dangling_behavior"] = "checkin"
    return config_dict


def run_migrations(config_dict):
    """Run all migrations."""
    logger.info(f"Running config migrations for {TO_VERSION}.")
    config_dict = add_dangling_behavior(config_dict)
    config_dict["_version"] = TO_VERSION
    return config_dict
